/*======================================================================*/
/*                                                                      */
/*      DDGI_DCCpp.c                                                    */
/*      Main module of DLL DDGI_DCCpp.                                  */
/*======================================================================*/
/*      J.P.PILLOU      13/03/2018                                     */
/*======================================================================*/
/*      Derived from DDGI_SPROG.dll                                      */
/*======================================================================*/
/*======================================================================*/

#include <windows.h>
#include <stdio.h>
#include "DDGI_Generic_def.h"
#include "DDGI_Generic_lib_def.h"
#include "dccpp_def.h"
#include "dccpp_extern.h"
#include "dccpp_util.h"
#include "dccpp_specific.h"
#include "dccpp_thread.h"

/*======================================================================*/
/*                              LOCAl FUNCTIONS                         */
/*======================================================================*/


BOOL    DCPI_InitDCCppContext(  HWND                    hWnd, 
                                HANDLE                  hCom );

int     DCPI_GetBaudRate(       HANDLE                  hCom );

BOOL    DCPI_CheckServer(       HWND                    hWnd,
                                HANDLE                  hCom );

BOOL    DCPI_TryServerConfig(   HWND                    hWnd, 
                                HANDLE                  hCom,
                                BOOL                    *pTransmitError );

BOOL    DCPI_ConfigureServer(   HWND                    hWnd, 
                                HANDLE                  hCom,
                                BOOL                    *pTransmitError );

void    DCPI_UpdateServer(      HWND                    hWnd, 
                                HANDLE                  hCom );

void    DCPI_PurgeBuffer(       HWND                    hWnd, 
                                HANDLE                  hCom );

BOOL    DCPI_SendAsciiBuffer(   HWND                    hWnd,
                                HANDLE                  hCom,
                                char                    *sCommandString,
                                BOOL                    bShowCommand );

BOOL    DCPI_ReadAsciiBuffer(   HWND                    hWnd,
                                HANDLE                  hCom,
                                unsigned char           xRcvBuffer[],
                                int                     *pNbChars,
                                BOOL                    bShowCommand );

BOOL    DCPI_GetAsciiVersionInfo( HWND                  hWnd,
                                HANDLE                  hCom );

BOOL    DCPI_ParseAsciiVersionInfo( HWND                hWnd, 
                                unsigned char           xRcvBuffer[],
                                char                    *sVersionString );

BOOL    DCPI_GetVersionInfo(    HWND                    hWnd,
                                HANDLE                  hCom );

void    DCPI_DisplayReadError(  HWND                    hWnd, 
                                RD_ERROR            nErrorCode,
                                char                    *sMsg );

void    DCPI_Tempo(             int                     nTempo );

/*======================================================================*/
/*                              GLOBAL VARIABLES                        */
/*======================================================================*/

/*======================================================================*/

BOOL WINAPI DllMain(    HINSTANCE hinstDLL,
                        DWORD     fdwReason,
                        LPVOID    lpvReserved )
{
  switch (fdwReason)
  {
  case DLL_PROCESS_ATTACH:
         /* Init Code here */
    break;

  case DLL_THREAD_ATTACH:
         /* Thread-specific init code here */
    break;

  case DLL_THREAD_DETACH:
        /* Thread-specific cleanup code here. */
    break;

  case DLL_PROCESS_DETACH:
        /* Cleanup code here */
    break;
  }
  /* The return value is used for successful DLL_PROCESS_ATTACH */
  return TRUE;
}

/*======================================================================*/
/*                      EXPORTABLE FUNCTION DEFINITIONS                 */
/*======================================================================*/

/*======================================================================*/
/*      DDGL_Register                                                   */

__declspec (dllexport) HANDLE
DDGL_Register(  BOOL            bEnglish,
                char            **ppDescString,
                char            **ppVersion,
                char            *sRailProtocols[],
                char            *sFeedbackProtocols[] )
{
  *ppDescString = DLL_NAME;
  *ppVersion = DLL_VERSION;
  sRailProtocols[0] = "DCC";
  sRailProtocols[1] = NULL;
  sFeedbackProtocols[0] = NULL;
}

/*======================================================================*/
/*      DDGL_StartServer                                                */

/*======================================================================*/
/*      DDGL_StartServer                                                */
/*      This function scans all COM ports of the computer, and          */
/*      looks for an P50X Interface attached.                           */
/*      In this first phase, all accesses (interface version, interface */
/*      speed, interface baud rate) are done synchronously (ie in the   */
/*      main thread which waits for responses).                         */

__declspec (dllexport) BOOL
DDGL_StartServer( DGI_SERVER_PARAMS  *pServerParams )
{
  HWND          hWnd;
  HANDLE        hCom = NULL;
  DCB           dcb;
  DWORD         dwError;
  BOOL          fSuccess;
  int           nComPort;
  char          sComPort[MAX_COM_PORT_NAME_LEN];
  BOOL          bFound = FALSE;
  DCC_COMPLETION_CODE   nCompletionCode;
  char          sMessage[128];

  hWnd = hMainWnd = pServerParams->hMainWindow;



  nComPort = 1;
  while ( nComPort <= MAX_COM_PORT_NUM && !bFound )
  {
    if ( nComPort < 10 )
      sprintf( sComPort, "COM%d", nComPort );
    else
      sprintf( sComPort, "\\\\.\\COM%d", nComPort );
     
    /*
    sprintf( sMessage, "Checking %s", sComPort );
    MessageBox( hWnd, sMessage, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    */
    
    hCom = CreateFile( 
        sComPort,
        GENERIC_READ | GENERIC_WRITE,
        0,    /* comm devices must be opened w/exclusive-access */
        NULL, /* no security attrs */
        OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
        0,    /* not overlapped I/O */
        NULL  /* hTemplate must be NULL for comm devices */ );

    if ( hCom != INVALID_HANDLE_VALUE)
    { 
      /*
      sprintf( sMessage, "COM%s connected", sComPort );
      MessageBox( hWnd, sMessage, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
      */

      bFound = TRUE;
      xDccCtxt.xServerDesc.nComPortNum = 
      pServerParams->nComPort = nComPort;
      pServerParams->hComPort = hCom;

      bFound = DCPI_CheckServer( pServerParams->hMainWindow,
                                 pServerParams->hComPort );

      if ( !bFound )
      {
        CloseHandle( hCom );
        hCom = INVALID_HANDLE_VALUE;
      }
      else 
      {
        xDccCtxt.xServerDesc.nComPortNum = nComPort;
        xDccCtxt.xServerDesc.nComPortBaudRate = 
                                DCPI_GetBaudRate( pServerParams->hComPort );
        xDccCtxt.hComPort = hCom;
        DCPU_SetTimeOut( hCom, DATA_COM_TIME_OUT );

        DCPI_UpdateServer( hWnd, hCom );
      }
    }

    if ( !bFound )
      nComPort++;
  }  /* end while */

  /* CloseHandle( hCom ); */
  if ( hCom != NULL && hCom != INVALID_HANDLE_VALUE )
  {
    /* create dump file */
    if ( pServerParams->bDumpFile )
    {
      pDumpFile = fopen( DUMP_FILE_NAME, "w" );
      fprintf( pDumpFile, "DLL %s %s\n", DLL_NAME, DLL_VERSION );
      fflush( pDumpFile );
    }
    pServerParams->hComPort = hCom;
    pServerParams->nComPort = nComPort;

    // uncomment following for generating debug file DBG_DCCpp.txt
    // pDebugFile = fopen( DEBUG_FILE_NAME, "w" );
  }
  else
  {
    pServerParams->hComPort = INVALID_HANDLE_VALUE;
    pServerParams->nComPort = 0;
  }


  return ( hCom != INVALID_HANDLE_VALUE ) ;
}

/*======================================================================*/
/*      DDGL_StopServer                                                 */
/*      This function stops the server attached to port hCom,           */
/*      and releases the pCom channel.                                  */

__declspec (dllexport) BOOL
DDGL_StopServer( DGI_SERVER_PARAMS *pServerParams )
{
  HWND                  hWnd;
  HANDLE                hCom;
  BOOL                  bEndWrite;

  hWnd = pServerParams->hMainWindow;
  hCom = pServerParams->hComPort;

  if ( pDumpFile != NULL )
  {
    DCPU_DumpCmdFifoToFile( pDumpFile );

    fprintf( pDumpFile, "\n===============================================\n" );
    fprintf( pDumpFile, "Total Nb of commands sent:             %ld\n"
                        "Total Nb of refresh commands sent:     %ld\n"
                        "Total Nb of local commands sent:       %ld\n"
                        "Total Nb of user commands sent:        %ld\n"
                        "Total Nb of errors:                    %ld\n"
                        "Total Nb of messages reveived:         %ld\n"
                        "Total Nb of Cmd Busy Msg:              %ld\n"
                        "Total Nb of overflows:                 %ld/%ld/%ld\n"
                        "Mean execution time (ticks):           %lf\n"
                        "Current Nb of active commands:         %d\n"
                        "Current Nb of pending commands:        %d\n"
                        "Current Nb of errors in error list:    %d\n"
                        "Current Nb of messages in msg list:    %d\n"
                        "Current Nb of free CmdCtxt:            %d\n",
        xDccCtxt.lTotalNbCmdSent,
        xDccCtxt.lTotalNbRefreshCmdSent,
        xDccCtxt.lTotalNbLocalCmdSent,
        xDccCtxt.lTotalNbUserCmdSent,
        xDccCtxt.lTotalNbErrors,
        xDccCtxt.lTotalNbMsgReceived,
        xDccCtxt.lTotalNbCmdBusyMsg,
        xDccCtxt.lNbSndOverflows,
        xDccCtxt.lNbErrOverflows,
        xDccCtxt.lNbSndOverflows,
        ( ( (double) xDccCtxt.lTotalDeltaTime ) / 
          ( (double) xDccCtxt.lTotalNbCmdSent ) ),
        xDccCtxt.xSndCmd.nNbCmd,
        xDccCtxt.xSndCmd.nNbPendingCmd,
        xDccCtxt.xError.nNbCmd,
        xDccCtxt.xRcvMsg.nNbCmd,
        xDccCtxt.nNbFreeCmdCtxt );
    fflush( pDumpFile );
    fclose( pDumpFile );
  }

  if ( pDebugFile != NULL )
    fclose( pDebugFile );

  if ( xDccCtxt.hIFThread )
  {
    bEndWrite = TerminateThread( xDccCtxt.hIFThread, 0 );
    DeleteCriticalSection( &xDccCtxt.xCriticalSection );

    if ( !bEndWrite )
     MessageBox( hWnd, "COULD NOT TERMINATE COMM THREAD", 
                 "XPRESSNET SERVER", 
                 MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    else
    {
      xDccCtxt.hIFThread = NULL;
    }
  }

  DCPU_FreeCommandCtxtEnv( &xDccCtxt.xSndCmd );
  DCPU_FreeCommandCtxtEnv( &xDccCtxt.xError );
  DCPU_FreeCommandCtxtEnv( &xDccCtxt.xRcvMsg );
      
  CloseHandle( hCom );
  hCom = NULL;

  return TRUE;
}


/*======================================================================*/
/*      DDGL_DisplayError                                               */

__declspec (dllexport) BOOL
DDGL_DisplayError( DGI_SERVER_PARAMS *pServerParams )
{
  HWND                  hWnd;
  HANDLE                hCom;
  char                  sMessage[MAX_MSG_LEN];

  hWnd = pServerParams->hMainWindow;
  hCom = pServerParams->hComPort;


  if ( xDccCtxt.xServerDesc.nLastRdErrorCode != RD_ERR_NONE )
  {
    strcpy( sMessage, 
            DCPU_RdErrorStr( xDccCtxt.xServerDesc.nLastRdErrorCode ) );
  }
  else
  {
    strcpy( sMessage, 
            DCPU_RcvCmdMnemonicStr( xDccCtxt.xServerDesc.nLastRdCommand ) );
  }

  MessageBox( hWnd, sMessage, "SPROG ERROR ", 
    MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
  xDccCtxt.xServerDesc.nLastRdErrorCode = RD_ERR_NONE;

  return TRUE;

}

/*======================================================================*/
/*      DDGL_DisplayServer                                              */
/*      This function displays current server characteristics           */

__declspec (dllexport) BOOL
DDGL_DisplayServer( DGI_SERVER_PARAMS *pServerParams )
{
  HWND                  hWnd;
  HANDLE                hCom;
  char                  sMessage[MAX_LONG_MSG_LEN];
  unsigned char         xBuf[6][12];

  hWnd = pServerParams->hMainWindow;
  hCom = pServerParams->hComPort;

  if ( hCom != NULL && hCom != INVALID_HANDLE_VALUE )
  {
    sprintf( sMessage,  
        "COM port____:\t\t%d\n"
        "Baud rate___:\t\t%d\n"
        "Version_____:\t\t%s\n",
        xDccCtxt.xServerDesc.nComPortNum, 
        xDccCtxt.xServerDesc.nComPortBaudRate,
        xDccCtxt.xServerDesc.sVersionString
        );
    MessageBox( hWnd, sMessage, "DCCpp SERVER", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    
  }
  return TRUE;
}

/*======================================================================*/
/*      DDGL_DisplayServerStatus                                        */
/*      This function displays current status of DCC protocol.          */

__declspec (dllexport) BOOL
DDGL_DisplayServerStatus( DGI_SERVER_PARAMS *pServerParams )
{
  HWND                  hWnd;
  HANDLE                hCom;
  char                  sMessage[2 * MAX_LONG_MSG_LEN];

  hWnd = pServerParams->hMainWindow;
  hCom = pServerParams->hComPort;

  if ( hCom != NULL && hCom != INVALID_HANDLE_VALUE )
  {
    sprintf( sMessage,  "Total Nb of commands sent:             %ld\n"
                        "Total Nb of refresh commands sent:     %ld\n"
                        "Total Nb of local commands sent:       %ld\n"
                        "Total Nb of user commands sent:        %ld\n"
                        "Total Nb of errors:                    %ld\n"
                        "Total Nb of overflows:                 %ld\n"
                        "Mean execution time (ticks):           %lf\n"
                        "Current Nb of active commands:         %d\n"
                        "Current Nb of pending commands:        %d\n"
                        "Current Nb of free CmdCtxt:            %d\n",
        xDccCtxt.lTotalNbCmdSent,
        xDccCtxt.lTotalNbRefreshCmdSent,
        xDccCtxt.lTotalNbLocalCmdSent,
        xDccCtxt.lTotalNbUserCmdSent,
        xDccCtxt.lTotalNbErrors,
        xDccCtxt.lNbSndOverflows,
        ( ( (double) xDccCtxt.lTotalDeltaTime ) / 
          ( (double) xDccCtxt.lTotalNbCmdSent ) ),
        xDccCtxt.xSndCmd.nNbCmd,
        xDccCtxt.xSndCmd.nNbPendingCmd,
        xDccCtxt.nNbFreeCmdCtxt );
    MessageBox( hWnd, sMessage, "DCC STATUS", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    
  }

  return TRUE;
}

/*======================================================================*/
/*      DDGL_EnableErrorDisplay                                         */
/*      This function enable global variable bDisplayErrors.            */

__declspec (dllexport) BOOL
DDGL_EnableErrorDisplay( DGI_SERVER_PARAMS *pServerParams )
{
  bDisplayErrors = TRUE;
  return TRUE;
}

/*======================================================================*/
/*      DDGL_DisplayStatistics                                          */
/*      This function displays current debug info on command handling   */
/*      (DCC context).                                                  */

__declspec (dllexport) BOOL
DDGL_DisplayStatistics( DGI_SERVER_PARAMS *pServerParams )
{
  HWND                  hWnd;
  HANDLE                hCom;
  char                  sMessage[2 * MAX_LONG_MSG_LEN];

  hWnd = pServerParams->hMainWindow;
  hCom = pServerParams->hComPort;

  if ( hCom != NULL && hCom != INVALID_HANDLE_VALUE )
  {
    sprintf( sMessage,  "Total Nb of commands sent:             %ld\n"
                        "Total Nb of refresh commands sent:     %ld\n"
                        "Total Nb of local commands sent:       %ld\n"
                        "Total Nb of user commands sent:        %ld\n"
                        "Total Nb of errors:                    %ld\n"
                        "Total Nb of messages reveived:         %ld\n"
                        "Total Nb of Cmd Busy Msg:              %ld\n"
                        "Total Nb of overflows:                 %ld/%ld/%ld\n"
                        "Mean execution time (ticks):           %lf\n"
                        "Current Nb of active commands:         %d\n"
                        "Current Nb of pending commands:        %d\n"
                        "Current Nb of errors in error list:    %d\n"
                        "Current Nb of messages in msg list:    %d\n"
                        "Current Nb of free CmdCtxt:            %d\n",
        xDccCtxt.lTotalNbCmdSent,
        xDccCtxt.lTotalNbRefreshCmdSent,
        xDccCtxt.lTotalNbLocalCmdSent,
        xDccCtxt.lTotalNbUserCmdSent,
        xDccCtxt.lTotalNbErrors,
        xDccCtxt.lTotalNbMsgReceived,
        xDccCtxt.lTotalNbCmdBusyMsg,
        xDccCtxt.lNbSndOverflows,
        xDccCtxt.lNbErrOverflows,
        xDccCtxt.lNbSndOverflows,
        ( ( (double) xDccCtxt.lTotalDeltaTime ) / 
          ( (double) xDccCtxt.lTotalNbCmdSent ) ),
        xDccCtxt.xSndCmd.nNbCmd,
        xDccCtxt.xSndCmd.nNbPendingCmd,
        xDccCtxt.xError.nNbCmd,
        xDccCtxt.xRcvMsg.nNbCmd,
        xDccCtxt.nNbFreeCmdCtxt );
    MessageBox( hWnd, sMessage, "DCC STATISTICS", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    
  }

  return TRUE;
}

/*======================================================================*/
/*      DDGL_DisplayRcvBuffer                                           */
/*      This function displays contents of last 256 received characters */

__declspec (dllexport) BOOL
DDGL_DisplayRcvBuffer( DGI_SERVER_PARAMS *pServerParams )
{
  DCPU_DumpDbgFifo( pServerParams->hMainWindow );
  return TRUE;
}

/*======================================================================*/
/*      DDGL_DisplayCmdFifo                                             */
/*      This function displays time tick and contents of last 32 cmds   */
/*      Snd + Rcv.                                                      */

__declspec (dllexport) BOOL
DDGL_DisplayCmdFifo( DGI_SERVER_PARAMS *pServerParams )
{
  DCPU_DumpCmdFifo( pServerParams->hMainWindow, NULL, FALSE );
  DCPU_DumpCmdFifoToFile( pDumpFile );
  return TRUE;
}

/*======================================================================*/
/*      DDGL_GetNextExpectedResponse                                    */
/*      Mapping of pGenericRcvData fields is done in                    */
/*      DCCT_GetNextExpectedResponse().                                 */

__declspec (dllexport) BOOL
DDGL_GetNextExpectedResponse(   pDGI_SERVER_PARAMS      pServerParams,
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData )
{
  long                  lTimeTick;
  BOOL                  bOK;

  lTimeTick = pGenericRcvData->lTimeTick;       /* unused */

  bOK = DCPT_GetNextExpectedResponse( pGenericRcvData );

  return bOK;
}

/*======================================================================*/
/*      DDGL_GetNextUnExpectedMessage                                   */
/*      Invoked on any time tick in order to get feedback messages.     */
/*      On any new time tick, schedules command XEvent, in order to     */
/*      check whether any new event has to be serviced.                 */


__declspec (dllexport) BOOL
DDGL_GetNextUnExpectedMessage(  pDGI_SERVER_PARAMS      pServerParams,
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData )
{
  long                  lTimeTick;
  CMD_DATA              xDummyDataArray[1]; 
  BOOL                  bOK;

  lTimeTick = pGenericRcvData->lTimeTick;

  if ( lTimeTick != xDccCtxt.xServerDesc.lCurTimeTick )
  {
    xDccCtxt.xServerDesc.lCurTimeTick = lTimeTick;
  }

  bOK = DCPT_GetNextUnExpectedMessage( pGenericRcvData );

  return bOK;
}

/*======================================================================*/
/*      DDGL_ResumeOperations                                           */
/*      Sends the RESUME OPERATIONS message to DCC network.             */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
   

__declspec (dllexport) BOOL
DDGL_ResumeOperations(  pDGI_SERVER_PARAMS      pServerParams,
                        pDGI_GENERIC_DATA       pGenericData )
{
  BOOL          bOK;

  /* xDccCtxt.xServerDesc.bEmergencyStop = FALSE; in DCPT_ExtraHandling */
  /* xDccCtxt.xServerDesc.bPowerOff = FALSE;            ""              */
  DCPS_RestartAllLocos();

  return 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                          pGenericData->pCmdTag, 
                         SND_RESUME_OPER_REQUEST, 0, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_3 );

}

/*======================================================================*/
/*      DDGL_PowerOff                                                   */
/*      Sends the POWER OFF (EMERGENCY OFF) message to DCC network.     */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*                                                                      */

__declspec (dllexport) BOOL
DDGL_PowerOff(          pDGI_SERVER_PARAMS      pServerParams,
                        pDGI_GENERIC_DATA       pGenericData )
{
  /* xDccCtxt.xServerDesc.bPowerOff = TRUE; set in DCpT_ExtraHandling */

  return 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_STOP_ALL_OPERS_REQUEST, 0, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_3 );
}

/*======================================================================*/
/*      DDGL_EmergencyStop                                              */
/*      Sends the EMERGENCY STOP message to DCC network.                */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*                                                                      */
/*      Performed only locally: no action forwarded to SPROG.           */

__declspec (dllexport) BOOL
DDGL_EmergencyStop(     pDGI_SERVER_PARAMS      pServerParams,
                        pDGI_GENERIC_DATA       pGenericData )
{
  /* xDccCtxt.xServerDesc.bEmergencyStop = TRUE; set in DCPT_ExtraHandling */
  DCPS_EmergencyStopAllLocos( pGenericData->pCmdTag );

 /* finally schedule emergency stop as standalone command to force feeback*/
 /* And pass CdmTag!!!.                                                 */
  return 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                          pGenericData->pCmdTag, 
                         SND_STOP_ALL_LOCOS_REQUEST, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_2 );
}

/*======================================================================*/
/*      DDGL_StopLoco                                                   */
/*      Send the emergency stop loco (immediate) to loco at address     */
/*      nAddress.                                                       */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*      - pGenericData->nAddress                                        */

__declspec (dllexport) BOOL
DDGL_StopLoco(          pDGI_SERVER_PARAMS      pServerParams,
                        pDGI_GENERIC_DATA       pGenericData )
{
  SND_CMD_MNEMONIC      nSendCommand;
  int                   nAddress;
  SPEED_STEP_MODE       nSpeedStepMode;
  int                   nSpeed;
  int                   nF0F12, nF0;
  int                   i, nCurIndex;
  int                   nNbAdBytes, nNbDataBytes;
  unsigned char         xAdByte[2], xDataByte[2];
  BOOL                  bOK;
  int                   nRegister;
  char                  sMessage[128];

  nAddress = pGenericData->nAddress;

  nRegister = DCPS_UpdateLocoParameters( nAddress, 0, SPEED_STEP_128,
                                        TRUE, 0, 0 );

  xDataArray[0] = nRegister;
  xDataArray[1] = nAddress;
  xDataArray[2] = 0;            // speed
  xDataArray[3] = 1;

  bOK =
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         NULL,
                         SND_LOCO_SPEED_128, 4, xDataArray, 
                         FALSE, CMD_LOCAL, PRI_3 );

  if ( bOK )
  {
    xDataArray[0] = nRegister;
    xDataArray[1] = nAddress;
    xDataArray[2] = -1;
    xDataArray[3] = 1;

    bOK =
      DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_LOCO_SPEED_128, 4, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_3 );
  }

  return bOK;

}

/*======================================================================*/
/*      DDGL_EnterServiceMode                                           */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*                                                                      */
/*      In DCCpp, this instruction does not have any impact.            */
/*      It is handled in standalone mode (no forwarding to base station.*/
/*      Merely returns RCV_ACK.                                         */

__declspec (dllexport) BOOL
DDGL_EnterServiceMode( pDGI_SERVER_PARAMS      pServerParams,
                       pDGI_GENERIC_DATA       pGenericData  )
{
  BOOL          bOK;

  bOK = 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag, 
                         SND_ENTER_SERVICE_MODE, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );
  return bOK;
}

/*======================================================================*/
/*      DDGL_DirectCVRead                                               */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*                                                                      */
/*      The implementation of this function is a bit special, since     */
/*      the DGI approach that the Direct CV read command does not       */
/*      return the result immediatly, but merely returns OK.            */
/*      The resukkt being handled by DDGL_ServiceModeResultRequest.     */
/*      But the SPROG returns the value after this function.            */
/*      The way to work this around is the following:                   */
/*      - declare SND_DIRECT_CV_READ as expecting response              */
/*        RCV_SRVC_MODE_RESP_DIRECT_CV (actually any response other     */
/*        than rCV_ACK), in order to store the numeric response.        */
/*      - invoke DCpT_PerformExtraHandling() for                        */
/*        * storing read value into global nCurrentCVValue.             */
/*        * replace nRdCommand by RCV_ACK, expected value.              */
/*                                                                      */
/*      nCurrentCVValue wil l be read and returned by                   */
/*      DDGL_ServiceModeResultRequest, which will invoke                */
/*      SND_SERVICE_MODE_EVENT_REQUEST in STANDALONE MODE.              */

__declspec (dllexport) BOOL
DDGL_DirectCVRead( pDGI_SERVER_PARAMS      pServerParams,
                   pDGI_GENERIC_DATA       pGenericData  )
{
  BOOL                  bOK;

  xDataArray[0] = pGenericData->nAddress;       // CV address
  xDataArray[1] = 1;    
  xDataArray[2] = 1;
   
  bOK = 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag, 
                         SND_DIRECT_CV_READ, 3, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_ServiceModeResultRequest                                   */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */

__declspec (dllexport) BOOL
DDGL_ServiceModeResultRequest( pDGI_SERVER_PARAMS      pServerParams,
                               pDGI_GENERIC_DATA       pGenericData  )
{
  return
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag, 
                         SND_SERVICE_MODE_EVENT_REQUEST, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );
}

/*======================================================================*/
/*      DDGL_DirectCVWrite                                              */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*      - pGenericData->nAddress  (CV number)                           */
/*      - pGenericData->nData[0]  (CV value)                            */

__declspec (dllexport) BOOL
DDGL_DirectCVWrite( pDGI_SERVER_PARAMS  pServerParams,
                    pDGI_GENERIC_DATA   pGenericData )
{
  BOOL                  bOK;

  xDataArray[0] = pGenericData->nAddress;       // CV address
  xDataArray[1] = pGenericData->nData[0];       // CV value
  xDataArray[2] = 1;    
  xDataArray[3] = 1;
   
  bOK = 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag, 
                         SND_DIRECT_CV_WRITE, 4, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_ExitServiceMode                                            */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */
/*                                                                      */
/*      In DCCpp, this instruction does not have any impact.            */
/*      It is handled in standalone mode (no forwarding to base station.*/
/*      Merely returns RCV_ACK.                                         */


__declspec (dllexport) BOOL
DDGL_ExitServiceMode( pDGI_SERVER_PARAMS      pServerParams,
                       pDGI_GENERIC_DATA       pGenericData  )
{
  BOOL          bOK;

  bOK = 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag, 
                         SND_EXIT_SERVICE_MODE, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );
  return bOK;
}

/*======================================================================*/
/*      DDGL_CommandStationVersionRequest                               */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */

__declspec (dllexport) BOOL
DDGL_CommandStationVersionRequest( pDGI_SERVER_PARAMS   pServerParams,
                                pDGI_GENERIC_DATA       pGenericData )
{
  return
   DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,  
                         SND_CMDSTATION_SOFT_V_REQUEST, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );
}

/*======================================================================*/
/*      DDGL_CommandStationStatusRequest                                */
/*                                                                      */
/*      This function is absolutely mandatory, for RUN initialization,  */
/*      Since this is the vehicle for DIF_ExecAndSchedule.              */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->xCmdTag                                         */


__declspec (dllexport) BOOL
DDGL_CommandStationStatusRequest( pDGI_SERVER_PARAMS    pServerParams,
                                pDGI_GENERIC_DATA       pGenericData )
{
 
  return
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_CMDSTATION_STATUS_REQUEST, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );
}


/*======================================================================*/
/*      DDGL_AccessoryInfoRequest                                       */
/*      Sent as CMD_STANDALONE, since processed at server level (not    */
/*      forwarded to SPROG.                                             */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      Address starts from 1 both in pCmdTag and in nAddress.          */
        

__declspec (dllexport) BOOL
DDGL_AccessoryInfoRequest( pDGI_SERVER_PARAMS      pServerParams,
                           pDGI_GENERIC_DATA       pGenericData )
{
  BOOL                  bOK = TRUE;
  int                   nAddress;

  nAddress = pGenericData->nAddress;    //  transmitted in pGenericData->pCmdTag

  bOK = 
      DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_ACCESSORY_INFO_REQUEST, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_DetectorInfoRequest                                        */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      Detector address starts from 1 both in pCmdTag and in nAddress. */
/*                                                                      */

__declspec (dllexport) BOOL
DDGL_DetectorInfoRequest(       pDGI_SERVER_PARAMS      pServerParams,
                                pDGI_GENERIC_DATA       pGenericData )
{
  BOOL                  bOK;
  int                   nAddress;

  nAddress = pGenericData->nAddress;    //  transmitted in pGenericData->pCmdTag

  bOK = DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_DETECTOR_INFO_REQUEST, 0, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_SetAccessoryState                                          */
/*      nAddress is the accessory address according to USER space,      */
/*      ie starting from 1 (NOT 0).                                     */
/*      The protocol address has the same convention (start from 1).    */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      - pGenericData->nData[0]:        bOut2                          */
/*      - pGenericData->nData[1]:        bOn                            */

__declspec (dllexport) BOOL
DDGL_SetAccessoryState(         pDGI_SERVER_PARAMS      pServerParams,
                                pDGI_GENERIC_DATA       pGenericData )
{
  int                   nAddress;
  BOOL                  bOut2, bOn;
  int                   i, nCurIndex;
  unsigned char         xAdDByte[2];
  BOOL                  bOK = TRUE;

  nAddress = pGenericData->nAddress;

  bOut2 = pGenericData->nData[0] != 0;
  bOn =   pGenericData->nData[1] != 0;

  if ( !DCPS_AccessoryDefined( nAddress ) )
  {
    xDataArray[0] = nAddress;                   // ID is address
    xDataArray[1] = (nAddress - 1) / 4;         // module address
    xDataArray[2] = ((nAddress - 1) % 4);       // submodule address

    // NOTE: SUBODUKE ADDRESSE STARTS FROM 1 in DCC++ NOT 0 !!! Add 1
    xDataArray[1]++;
    //===============

    bOK =
      DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                           NULL,
                           SND_DEFINE_TURNOUT, 3, xDataArray, 
                           FALSE, CMD_LOCAL, PRI_1 );
    if ( bOK )
    {
      bOK =
        DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                           NULL,
                           SND_STORE_IN_EEPROM, 0, xDataArray, 
                           FALSE, CMD_LOCAL, PRI_0 );
    }
  }

  if ( bOK )
  {
    if ( bOn )  // activate
    {
      xDataArray[0] = nAddress;                 // ID
      xDataArray[1] = bOut2;                   // 1 if THROWN 
      DCPS_SetAccessoryState( nAddress, bOut2 );

      bOK =
        DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                           pGenericData->pCmdTag,
                           SND_ACCESSORY_OPER, 2, xDataArray, 
                           FALSE, CMD_NORMAL, PRI_0 );
    }
    else        // release: juste send ACK in standalone mode
    {
      bOK =
        DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                           pGenericData->pCmdTag,
                           SND_ACCESSORY_OPER, 0, xDataArray, 
                           FALSE, CMD_STANDALONE, PRI_0 );
    }
  }

  return bOK;
}

/*======================================================================*/
/*      DDGL_LocoInfoRequest                                            */
/*      Request speed, and function info for loco at address nAddress   */
/*      Sent as CMD_STANDALONE, since processed at server level (not    */
/*      forwarded to SPROG.                                             */
/*                                                                      */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*                                                                      */
/*      UNIMPLENTED IN SPROG, since no other source                     */
/*      If later on it had to be implemented, processing should be      */
/*      CMD_STANDALONE, and  address retrieved from pCmdTag.            */

/*
__declspec (dllexport) BOOL
DDGL_LocoInfoRequest(           pDGI_SERVER_PARAMS      pServerParams,
                                pDGI_GENERIC_DATA       pGenericData )
{
  SND_CMD_MNEMONIC      nSendCommand;
  BOOL                  bOK;

   bOK = 
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_LOCO_INFO_REQUEST, xDataArray, 
                         FALSE, CMD_STANDALONE, PRI_0 );

  return bOK;
}
*/

/*======================================================================*/
/*      DDGL_SetSpeed                                                   */
/*      Sets loco speed and direction in XpressNet V3.                  */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      - pGenericData->nData[0]:        nSpeedStepMode                 */
/*      - pGenericData->nData[1]:        nSpeed                         */
/*      - pGenericData->nData[2]:        nF0F12                         */
/*      - pGenericData->nData[3]:        bForward                       */
/*                                                                      */

__declspec (dllexport) BOOL
DDGL_SetSpeed(                  pDGI_SERVER_PARAMS      pServerParams,
                                pDGI_GENERIC_DATA       pGenericData )
{
  SND_CMD_MNEMONIC      nSendCommand;
  int                   nAddress;
  int                   nSpeed;
  BOOL                  bForward;
  BOOL                  bOK;
  int                   nRegister;      // register number, NOT index

  nAddress = pGenericData->nAddress;
  nSpeed = (SPEED_STEP_MODE) pGenericData->nData[1];
  if ( nSpeed < 0 )
    nSpeed = -nSpeed;   // absolute value
  bForward = pGenericData->nData[3];

  nRegister = DCPS_UpdateLocoParameters( nAddress, nSpeed, SPEED_STEP_128,
                                        bForward, 0, 0 );

  xDataArray[0] = nRegister;
  xDataArray[1] = nAddress;
  xDataArray[2] = nSpeed;
  xDataArray[3] = bForward? 1:0;

  bOK =
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         SND_LOCO_SPEED_128, 4, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_2 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_SetF0F4                                                    */
/*      Sets all functions F0 to F4.                                    */
/*      nFuncValue contains the 5 bits of functions F0-F4 ordered       */
/*      from LSB (F0) to MSB (F4).                                      */
/*                                                                      */
/*      speed passed, but normally unused.                              */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      - pGenericData->nData[0]:        nFuncValue (bits 12-0)         */
/*      - pGenericData->nData[1]:        nSpeed                         */
/*                                                                      */

__declspec (dllexport) BOOL
DDGL_SetF0F4( pDGI_SERVER_PARAMS        pServerParams,
              pDGI_GENERIC_DATA         pGenericData )
{
  SND_CMD_MNEMONIC      nSendCommand;
  int                   nAddress;
  int                   nFuncValue;
  BOOL                  bOK;

  nAddress = pGenericData->nAddress;
  nFuncValue = pGenericData->nData[0];
  nSendCommand = SND_FUNCTION_GROUP_1;

  DCPS_UpdateLocoFx( nAddress, nFuncValue, (unsigned char) 0x1F );

  xDataArray[0] = nAddress;
  xDataArray[1] = DCPS_SetFxByte( nFuncValue, (unsigned int) 0x1F );

  bOK =
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         nSendCommand, 2, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_SetF5F8                                                    */
/*      Sets functions F5 to F8 on loco at address nAddress             */
/*      nFuncValue contains the 4 bits of functions F5-F8 on bits 5-8.  */
/*                                                                      */
/*      speed passed, but normally unused.                              */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      - pGenericData->nData[0]:        nFuncValue (bits 12-0)         */
/*      - pGenericData->nData[1]:        nSpeed                         */
/*                                                                      */

__declspec (dllexport) BOOL
DDGL_SetF5F8( pDGI_SERVER_PARAMS        pServerParams,
              pDGI_GENERIC_DATA         pGenericData  )
{
  SND_CMD_MNEMONIC      nSendCommand;
  int                   nAddress;
  int                   nFuncValue;
  BOOL                  bOK;

  nAddress = pGenericData->nAddress;
  nFuncValue = pGenericData->nData[0];
  nSendCommand = SND_FUNCTION_GROUP_2;

  DCPS_UpdateLocoFx( nAddress, nFuncValue, (unsigned char) 0x01E0 );

  xDataArray[0] = nAddress;
  xDataArray[1] = DCPS_SetFxByte( nFuncValue, (unsigned int) 0x01E0 );

  bOK =
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         nSendCommand, 2, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*      DDGL_SetF9F12                                                   */
/*      Sets functions F9 to F12 on loco at address nAddress            */
/*      nFuncValue contains the 4 bits of functions F9-F12 on bits 9-12.*/
/*                                                                      */
/*      speed passed, but normally unused.                              */
/*                                                                      */
/*      Field usage                                                     */
/*      - pGenericData->pCmdTag                                         */
/*      - pGenericData->nAddress                                        */
/*      - pGenericData->nData[0]:        nFuncValue (bits 12-0)         */
/*      - pGenericData->nData[1]:        nSpeed                         */
/*                                                                      */

__declspec (dllexport) BOOL
DDGL_SetF9F12( pDGI_SERVER_PARAMS        pServerParams,
               pDGI_GENERIC_DATA         pGenericData  )
{
  SND_CMD_MNEMONIC      nSendCommand;
  int                   nAddress;
  int                   nFuncValue;
  BOOL                  bOK;

  nAddress = pGenericData->nAddress;
  nFuncValue = pGenericData->nData[0];
  nSendCommand = SND_FUNCTION_GROUP_3;

  DCPS_UpdateLocoFx( nAddress, nFuncValue, (unsigned char) 0x1E00 );

  xDataArray[0] = nAddress;
  xDataArray[1] = DCPS_SetFxByte( nFuncValue, (unsigned int) 0x1E00 );

  bOK =
    DCPT_ScheduleSndMsg( pServerParams->hMainWindow, 
                         pGenericData->pCmdTag,
                         nSendCommand, 2, xDataArray, 
                         FALSE, CMD_NORMAL, PRI_0 );

  return bOK;
}

/*======================================================================*/
/*                              INTERNAL FUNCTIONS                      */
/*======================================================================*/

/*======================================================================*/
/*      DCPI_InitDCCppContext                                           */
/*      This function performs all initialization related to interface  */
/*      with DCC protocol.                                              */
/*      - creates write and read threads.                               */
/*      - initializes critical section.                                 */

BOOL
DCPI_InitDCCppContext( HWND hWnd, HANDLE hCom )
{
  BOOL                  bOK = TRUE;

  xDccCtxt.hIFThread = 
    CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE) DCPT_IFThreadFunc, 
                  (LPVOID) hWnd, 0, &xDccCtxt.lIFThreadID );
  if ( xDccCtxt.hIFThread == NULL )
  {
    MessageBox( hWnd, "COULD NOT CREATE READ or WRITE THREAD", 
                "SPROG SERVER", 
                MB_APPLMODAL | MB_OK | MB_ICONERROR );
    bOK = FALSE;
  }

  if ( bOK )
  {
    InitializeCriticalSection( &xDccCtxt.xCriticalSection );

    xDccCtxt.lTotalNbCmdSent            = 0L;
    xDccCtxt.lTotalNbUserCmdSent        = 0L;
    xDccCtxt.lTotalNbErrors             = 0L;
    xDccCtxt.lTotalNbMsgReceived        = 0L;
    xDccCtxt.lTotalNbCmdBusyMsg         = 0L;
    xDccCtxt.lTotalDeltaTime            = 0L;
    xDccCtxt.lNbSndOverflows            = 0L;
    xDccCtxt.lNbErrOverflows            = 0L;
    xDccCtxt.lNbRcvOverflows            = 0L;

    /* free element list maintained !!! */
    /* xDccCtxt.pFreeCmdCtxtList   = NULL; */
    /* xDccCtxt.nNbFreeCmdCtxt     = 0;    */

    DCPU_FreeCommandCtxtEnv( &xDccCtxt.xSndCmd );
    DCPU_FreeCommandCtxtEnv( &xDccCtxt.xError );
    DCPU_FreeCommandCtxtEnv( &xDccCtxt.xRcvMsg );

    xDccCtxt.nRcvMsgWrPtr = 0;
    xDccCtxt.nRcvMsgRdPtr = 0;
  }
  DCPU_SetTimeOut( hCom, DATA_COM_TIME_OUT );

  return bOK;
}

/*======================================================================*/
/*      DCPI_GetBaudRate                                                */
/*      This function returns the baud rate of COM port hCom            */

int
DCPI_GetBaudRate( HANDLE hCom )
{
  DCB                   xDcb;
  int                   nBaudRate;

  if ( !GetCommState( hCom, &xDcb ) )
    nBaudRate = 0;
  else
    nBaudRate = xDcb.BaudRate;

  return nBaudRate;
}

/*======================================================================*/
/*      DCPI_CheckServer                                                */
/*      This function:                                                  */
/*      - saves current DCB of hCom.                                    */
/*      - Then sets DCB to 2400, 4800, 9600, 19200, 38400               */
/*        and checks whether the interface answers.                     */

BOOL
DCPI_CheckServer( HWND hWnd, HANDLE hCom  )
{
  DCB                   xDcb, xDcbSave;
  COMMTIMEOUTS          xTimeOut, xTimeOutSave;
  BOOL                  bOK = FALSE;
  BOOL                  bCommStatusOK = FALSE;
  BOOL                  bModified = FALSE;
  BOOL                  bTransmitError;
  char                  sMessage[MAX_MSG_LEN];

  if ( !GetCommState( hCom, &xDcb ) )
    bCommStatusOK = FALSE;
  else
  {
    bCommStatusOK = TRUE;
    xDcbSave = xDcb;
    GetCommTimeouts( hCom, &xTimeOut );
    xTimeOutSave = xTimeOut;
    bModified = TRUE;
 
    xDcb.DCBlength      = sizeof(DCB);
    xDcb.ByteSize       = 8;
    xDcb.Parity         = NOPARITY;
    /* xDcb.StopBits       = ONESTOPBIT; */
    xDcb.StopBits       = TWOSTOPBITS;
    xDcb.fOutxCtsFlow   = TRUE;
    xDcb.fDtrControl    = DTR_CONTROL_DISABLE;
    xDcb.fBinary        = TRUE;
    xDcb.fOutxDsrFlow   = FALSE;
    /* RTS CONTROL: CRITICAL!! severe program hang up if not set as follows!*/
    xDcb.fRtsControl    = RTS_CONTROL_ENABLE;  /* CRITICAL!!!: */
    xDcb.fParity        = FALSE;
    xDcb.fDsrSensitivity = FALSE;
    xDcb.fOutX          = FALSE;
    xDcb.fInX           = FALSE;
    xDcb.fErrorChar     = FALSE;
    xDcb.fNull          = FALSE;

    DCPU_SetTimeOut( hCom, INIT_PHASE_TIME_OUT );

    xDcb.BaudRate = DCCPP_BAUD_RATE;
    SetCommState( hCom, &xDcb );

    DCPI_Tempo( 200 );
    SetCommBreak( hCom );
    DCPI_Tempo( 1000 );           /* 2 seconds */
    ClearCommBreak( hCom );
    DCPI_Tempo( 200 );

    bOK = DCPI_TryServerConfig( hWnd, hCom, &bTransmitError );

    if ( bModified && !bOK )   /* restore initial state */
    {
      SetCommState( hCom, &xDcbSave );
      SetCommTimeouts( hCom, &xTimeOutSave );
    }
  }

  return bOK;
}

/*======================================================================*/
/*      DCPI_TryServerConfig                                            */
/*      This function sends the status request command <s>, in order to */
/*      retrieve the DCCpp version string.                              */

BOOL
DCPI_TryServerConfig( HWND hWnd, HANDLE hCom, BOOL *pTransmitError )
{
  BOOL                  bOK = FALSE;
  int                   i;
  int                   nNbChars;

  bOK = DCPI_GetAsciiVersionInfo( hWnd, hCom );
  // and try a second time
  if ( !bOK )
    bOK = DCPI_GetAsciiVersionInfo( hWnd, hCom );

  return bOK;
}

/*======================================================================*/
/*      DCPI_GetAsciiVersionInfo                                        */
/*      Retrieves DCCpp version info by sending command <s>             */
/*      Since DCPI_ReadAsciiBuffer reads until the closing bracket '>'  */
/*      is found, it must be invoked twice.                             */

BOOL    
DCPI_GetAsciiVersionInfo( HWND hWnd, HANDLE hCom )
{
  unsigned char         xRcvBuffer[MAX_DCCPP_CMD_LEN];
  int                   nNbChars;
  BOOL                  bOK;
  char                  sMessage[MAX_STR_LEN];

  if ( !DCPI_SendAsciiBuffer( hWnd, hCom, "<s>", FALSE ) )
  {
    MessageBox( hWnd, "CANNOT WRITE <s> command", "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
    bOK = FALSE;
  }
  else
  {
    if ( !DCPI_ReadAsciiBuffer( hWnd, hCom, xRcvBuffer, &nNbChars, FALSE ) )
    {
      MessageBox( hWnd, "CANNOT READ command response #1", "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
      bOK = FALSE;
      xRcvBuffer[nNbChars] = '\0';
      sprintf( sMessage, "%d chars" , nNbChars );
      MessageBox( hWnd, sMessage, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
      MessageBox( hWnd, xRcvBuffer, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION ); // <p0> displayed
    }
    else
    {
      /*
      sprintf( sMessage, "%d chars" , nNbChars );
      MessageBox( hWnd, sMessage, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
      MessageBox( hWnd, xRcvBuffer, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION ); // <p0> displayed
      */

      if ( !DCPI_ReadAsciiBuffer( hWnd, hCom, xRcvBuffer, &nNbChars, FALSE ) )
      {
        MessageBox( hWnd, "CANNOT READ command response #2", "", 
          MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
        xRcvBuffer[nNbChars] = '\0';

        sprintf( sMessage, "%d chars" , nNbChars );
        MessageBox( hWnd, sMessage, "", 
          MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
        MessageBox( hWnd, xRcvBuffer, "", 
          MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );

      }
      else
      {
        /*
        sprintf( sMessage, "%d chars" , nNbChars );
        MessageBox( hWnd, sMessage, "", 
          MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
        MessageBox( hWnd, xRcvBuffer, "", 
          MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
        */

       }

      bOK = DCPI_ParseAsciiVersionInfo( hWnd, xRcvBuffer, 
                                        xDccCtxt.xServerDesc.sVersionString );

      // and repeat read of next <...> until exhausted (purge buffer)
      while ( DCPI_ReadAsciiBuffer( hWnd, hCom, xRcvBuffer, &nNbChars, FALSE ) )
      {
        /* for debug 
        xRcvBuffer[nNbChars] = '\0';
        MessageBox( hWnd, xRcvBuffer, "", MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
        */
      }
        

    }
  }
  return bOK;
}

/*======================================================================*/
/*      DCPI_ParseAsciiVersionInfo                                      */
/*      This function extracts version string between ':' and '/'       */

BOOL
DCPI_ParseAsciiVersionInfo( HWND        hWnd, 
                        unsigned char   xRcvBuffer[], 
                        char            *sVersionString )
{
  int                   nStartIndex, nStopIndex;
  unsigned char         xExpectedChar1[] = { ':', '\0' };
  unsigned char         xExpectedChar2[] = { '/', '\0' };
  unsigned char         cSaveChar;
  BOOL                  bOK = FALSE;

  nStartIndex = 0;
  bOK = DCPU_GetNextCharIndex( xRcvBuffer, xExpectedChar1, &nStartIndex );
  nStartIndex++;        // next char

  /* Version */
  if ( bOK )
  {
    nStopIndex = nStartIndex;
    if ( DCPU_GetNextCharIndex( xRcvBuffer, xExpectedChar2, &nStopIndex ) )
    {
      cSaveChar = xRcvBuffer[nStopIndex];
      xRcvBuffer[nStopIndex] = '\0';
      strcpy( sVersionString, &xRcvBuffer[nStartIndex] );
      xRcvBuffer[nStopIndex] = cSaveChar;
    }
    else
      bOK = FALSE; 
  }
     
  return bOK;
}



/*======================================================================*/
/*      DCPI_UpdateServer                                               */
/*      This function is invoked, once the server is started,           */
/*      It invokes DCPI_InitDCCppContext, which performs the following  */
/*      server updates:                                                 */
/*      - creates write and read threads.                               */
/*      - initializes critical section.                                 */
/*      - intializes 
/*      This is still part of the init phase.                           */

void
DCPI_UpdateServer( HWND hWnd, HANDLE hCom )
{
  DCPI_InitDCCppContext( hWnd, hCom ); 
  DCPS_InitS88();
}

/*======================================================================*/
/*      DCPI_PurgeBuffer                                                */
/*      This function is invoked on opening the server, to  read all    */
/*      spurious commands accumulated in the interface  buffer.         */
/*      When IF buffer is empty, the read error code should be          */
/*      RD_ERR_NO_HEADER.                                           */

void
DCPI_PurgeBuffer( HWND hWnd, HANDLE hCom )
{
  unsigned char         xRcvByte[256];
  DWORD                 nNbBytesRead;
  
  ReadFile( hCom, &xRcvByte, 256, &nNbBytesRead, NULL );
}

/*======================================================================*/
/*      DCPI_SendAsciiBuffer                                            */
/*      - appends a CR to command string sCmdSgtring, and stores it     */
/*      into xSendArray.                                                */
/*      - write full string + CR to hCom.                               */
/*                                                                      */
/*      RETURN: TRUE if number of bytes written is what is should be.   */
/*              FALSE otherwise.                                        */

BOOL
DCPI_SendAsciiBuffer( HWND              hWnd,
                HANDLE                  hCom,
                char                    *sCmdString,
                BOOL                    bShowCommand )
{
  int                   nLen;
  DWORD                 nNbBytesWritten;
  unsigned char         xSendArray[MAX_DCCPP_CMD_LEN];
  BOOL                  bOK = TRUE;
  char                  sMessage[128];

  strcpy( xSendArray, sCmdString );
  nLen = strlen( xSendArray );
  xSendArray[nLen] = (unsigned char) 0x0D;
  xSendArray[nLen + 1] = '\0';
  nLen++;

  if ( !WriteFile( hCom, &xSendArray, nLen, &nNbBytesWritten, NULL ) ||
       nNbBytesWritten != nLen )
    bOK = FALSE;

  if ( bShowCommand )
  {
    sprintf( sMessage, 
      "WRITE %d bytes: %02X %02X %02X %02X %02X %02X %02X %02X ",
        nLen,
        xSendArray[0], xSendArray[1], xSendArray[2], xSendArray[3], 
        xSendArray[4], xSendArray[5], xSendArray[6], xSendArray[7] ); 
      MessageBox(hWnd, sMessage, "", MB_APPLMODAL | MB_OK | MB_ICONINFORMATION);
  }

  return bOK;
}

/*======================================================================*/
/*      DCPI_ReadAsciiBuffer                                            */
/*      Reads receive buffer until character '>' is found.              */
/*      In this early phase (INIT), any read reads one single byte!!!   */
/*                                                                      */
/*      RETURN: TRUE if response received, FALSE if not.                */

BOOL
DCPI_ReadAsciiBuffer( HWND              hWnd,
                HANDLE                  hCom,
                unsigned char           xRcvBuffer[],
                int                     *pNbChars,
                BOOL                    bShowCommand )
{
  BOOL                  bEnd = FALSE;
  BOOL                  bOK = FALSE;
  DWORD                 nNbBytesRead;
  int                   nRcvIndex;
  int                   i;
  char                  sMessage[128];
  char                  sMessage2[128];

  for ( i = 0; i < MAX_DCCPP_CMD_LEN; i++ )
    xRcvBuffer[i] = '\0';
  
  nRcvIndex = 0;
  while ( !bEnd )
  {
    nNbBytesRead = 0;
    ReadFile( hCom, &xRcvBuffer[nRcvIndex], 1, &nNbBytesRead, NULL );
    if ( nNbBytesRead < 1 )
    {
      bEnd = TRUE;              /* time out !! */
      *pNbChars = nRcvIndex;
      bOK = FALSE;
    } 
    else
    {
      if ( xRcvBuffer[nRcvIndex] == '>' )
      {
        bOK = TRUE;
        bEnd = TRUE;
      }
      nRcvIndex++;
    }
  }

  *pNbChars = nRcvIndex;

  if ( bShowCommand )
  {
    if ( bOK )
    {
      sprintf( sMessage, 
        "READ OK %d bytes: %02X %02X %02X %02X %02X %02X %02X %02X ",
        *pNbChars,
        xRcvBuffer[0], xRcvBuffer[1], xRcvBuffer[2], xRcvBuffer[3], 
        xRcvBuffer[4], xRcvBuffer[5], xRcvBuffer[6], xRcvBuffer[7] ); 
      sprintf( sMessage2, 
        "%c %c %c %c %c %c %c %c ",
        xRcvBuffer[0], xRcvBuffer[1], xRcvBuffer[2], xRcvBuffer[3], 
        xRcvBuffer[4], xRcvBuffer[5], xRcvBuffer[6], xRcvBuffer[7] ); 
      MessageBox(hWnd, sMessage, sMessage2, MB_APPLMODAL | MB_OK | MB_ICONINFORMATION);
    }
    else
    {
      sprintf( sMessage, 
        "READ ERROR %d bytes: %02X %02X %02X %02X %02X %02X %02X %02X ",
        *pNbChars,
        xRcvBuffer[0], xRcvBuffer[1], xRcvBuffer[2], xRcvBuffer[3], 
        xRcvBuffer[4], xRcvBuffer[5], xRcvBuffer[6], xRcvBuffer[7] ); 
      sprintf( sMessage2, 
        "%c %c %c %c %c %c %c %c ",
        xRcvBuffer[0], xRcvBuffer[1], xRcvBuffer[2], xRcvBuffer[3], 
        xRcvBuffer[4], xRcvBuffer[5], xRcvBuffer[6], xRcvBuffer[7] ); 
       MessageBox(hWnd, sMessage, sMessage2, MB_APPLMODAL | MB_OK | MB_ICONINFORMATION);
    }
  }

  return bOK;
}

/*======================================================================*/
/*      DCPI_DisplayReadError                                           */
/*      Displays a message window with explicit string                  */
/*      describing erros which occured on read command.                 */

void
DCPI_DisplayReadError( HWND hWnd, RD_ERROR nErrorCode, char *sMsg )
{
  char          sMessage[MAX_MSG_LEN];

  if ( bDisplayErrors )
  {
    sprintf( sMessage, "READ ERROR %s: %s (%d)", 
        sMsg, DCPU_RdErrorStr( nErrorCode ), nErrorCode );
    MessageBox( hWnd, sMessage, "", MB_APPLMODAL | MB_OK | MB_ICONERROR );
  }
}

/*======================================================================*/
/*      DCPI_Tempo                                                      */
/*      Returns after nTempo (expressed in milliseconds).               */

void
DCPI_Tempo( int nTempo )
{
  SYSTEMTIME            xStartTime;
  SYSTEMTIME            xCurrentTime;
  long                  lStartTime, lCurrentTime;
  BOOL                  bEnd = FALSE;

  GetSystemTime( &xStartTime );
  lStartTime = ( 60 * xStartTime.wMinute + xStartTime.wSecond ) * 1000 
                + xStartTime.wMilliseconds;

  while ( !bEnd )
  {
    GetSystemTime( &xCurrentTime );
    lCurrentTime = ( 60 * xCurrentTime.wMinute + xCurrentTime.wSecond ) * 1000 
                   + xCurrentTime.wMilliseconds;
    if ( xCurrentTime.wHour != xStartTime.wHour )       /* hour has changed! */
      lCurrentTime += 3600000;

    if ( lCurrentTime - lStartTime >= nTempo )
      bEnd = TRUE;
  }

}

