/*======================================================================*/
/*                                                                      */
/*      DDGI_Generic_def.h                                              */
/*      Generic definition file common to ALL DDGI DLLs                 */
/*======================================================================*/
/*      J.P.PILLOU      25/08/2012                                      */
/*======================================================================*/ 
/*      Modified        27/02/2013                                      */
/*      Fields added in structure DGI_SERVER_PARAMS                     */
/*        DGI_LANGUAGE  nLanguage;                                      */
/*        int           nSpare[DGI_NB_SRV_PARAM_SPARES];                */
/*        char          *sExecDirectory;                                */
/*======================================================================*/  

#define DGI_NB_SRV_PARAM_SPARES         8

#define DGI_MAX_DATA_WORDS              5
#define DGI_MAX_DATA_ITEMS              8

/* command station status */
#define DGI_CS_STATUS_STOP_BIT          0
#define DGI_CS_STATUS_POWER_OFF_BIT     1
#define DGI_CS_STATUS_START_MODE_BIT    2
#define DGI_CS_STATUS_SERVICE_MODE_BIT  3
#define DGI_CS_STATUS_OVERHEAT_BIT      4
#define DGI_CS_STATUS_POWER_UP_BIT      6
#define DGI_CS_STATUS_RAM_CHK_ERR_BIT   7

typedef enum
{ 
  DGI_L_FRENCH,
  DGI_L_ENGLISH
}       DGI_LANGUAGE;

                                                       
typedef enum
{
  DGI_F_START_SERVER,
  DGI_F_STOP_SERVER,
  DGI_F_DISPLAY_SERVER,
  DGI_F_DISPLAY_ERROR,
  DGI_F_DISPLAY_SERVER_STATUS,
  DGI_F_ENABLE_ERROR_DISPLAY,
  DGI_F_DISPLAY_DCC_STATISTICS,
  DGI_F_DISPLAY_RCV_BUFFER,
  DGI_F_DISPLAY_CMD_FIFO,
  DGI_F_GET_NEXT_EXPECTED_RESPONSE,
  DGI_F_GET_NEXT_UNEXPECTED_MESSAGE,
  DGI_F_SCAN_RCV_MESSAGES,
  DGI_F_GET_PROTOCOL_TYPE,
  DGI_F_RESUME_OPERATIONS,
  DGI_F_POWER_OFF,
  DGI_F_EMERGENCY_STOP,
  DGI_F_STOP_LOCO,
  DGI_F_ENTER_SERVICE_MODE,
  DGI_F_DIRECT_CV_READ,
  DGI_F_S_MODE_RESULT_REQUEST,
  DGI_F_DIRECT_CV_WRITE,
  DGI_F_EXIT_SERVICE_MODE,
  DGI_F_CS_VERSION_REQUEST,
  DGI_F_CS_STATUS_REQUEST,
  DGI_F_ACCESSORY_INFO_REQUEST,
  DGI_F_DETECTOR_INFO_REQUEST,
  DGI_F_SET_ACCESSORY_STATE,
  DGI_F_LOCO_INFO_REQUEST,
  DGI_F_SET_SPEED,
  DGI_F_SET_F0F12,
  DGI_F_SET_F0F4,
  DGI_F_SET_F5F8,
  DGI_F_SET_F9F12,
  DGI_F_OPERATION_MODE_PROG,
}               DGI_LIB_FUNC_INDEX;

typedef enum 
{  
  DGI_COMP_NONE,
  DGI_COMP_OK,
  DGI_COMP_OK_RD,
  DGI_COMP_READ_ERR,
  DGI_COMP_PROT_ERR,
  DGI_COMP_PROG_ERR,
  DGI_COMP_MUDH_ERR,
  DGI_COMP_OPER_ERR,
  DGI_COMP_SEND_ERR,
  DGI_COMP_LOCAL_OK
}               DCC_COMPLETION_CODE;

typedef enum
{
  SPEED_STEP_14,
  SPEED_STEP_27,
  SPEED_STEP_28,
  SPEED_STEP_128
}               SPEED_STEP_MODE;

typedef enum
{  RCV_PC_IF_ERROR,                             /* 0 */
  RCV_IF_CS_ERROR,
  RCV_UNKNOWN_COMM_ERROR,
  RCV_ACK,
  RCV_NO_CS_TIME_SLOT,
  RCV_IF_BUFFER_OVERFLOW,
  RCV_COMMAND_NOT_UNDERSTOOD,
  RCV_NORMAL_OPER_RESUMED,
  RCV_TRACK_POWER_OFF,
  RCV_EMERGENCY_STOP,
  RCV_SERVICE_MODE_ENTRY,                       /* 10 */
  RCV_FEEDBACK_BROADCAST,
  RCV_PROG_SHORT_CIRCUIT,
  RCV_PROG_DATA_BYTE_NOT_FOUND,
  RCV_PROG_CMDSTATION_BUSY,
  RCV_PROG_CMDSTATION_READY,
  RCV_SRVC_MODE_RESP_REG_PAGED,
  RCV_SRVC_MODE_RESP_DIRECT_CV,
  RCV_CMDSTATION_SOFT_V_ANSWER_XBUS,
  RCV_CMDSTATION_SOFT_V_ANSWER_XPNET,
  RCV_CMDSTATION_STATUS,                        /* 20 */
  RCV_TRANSFER_ERRORS,                          
  RCV_CMDSTATION_BUSY,
  RCV_UNSUPPORTED_COMMAND,
  RCV_ACCESSORY_INFO,
  RCV_DETECTOR_INFO,
  RCV_NORMAL_LOCO_INFO,
  RCV_MULTI_UNIT_LOCO_INFO,
  RCV_MULTI_UNIT_AD_LOCO_INFO,
  RCV_DOUBLE_HEADER_LOCO_INFO,
  RCV_LOCO_OPER_BY_ANOTHER_DEVICE,              /* 30 */
  RCV_FUNCTION_STATUS,
  RCV_LOCO_AD_RETRIEVAL,
  RCV_DOUBLE_HEADER_ERROR,
  RCV_MU_DH_ERROR,                              /* 34 */

  RCV_LENZ_IF_V_ANSWER,
  RCV_LENZ_IF_SET_ADDRESS_ACK,
  RCV_LENZ_IF_SET_BD_RATE_ACK,

  RCV_EVENT_REPORT,                             /* 38 */
  RCV_LOCO_EVENT_REPORT,
  RCV_DETECTOR_EVENT_REPORT,
  RCV_TURNOUT_EVENT_REPORT,
  RCV_STATUS_EVENT_REPORT,
  RCV_PRE_POWER_OFF_EVENT,

  RCV_NB_DETECTORS,                             /* 44 */
  RCV_DETECTOR_INITIAL_STATES,
  RCV_DETECTOR_STATES,                          /* 46 */

  // DCC++ feedbacks
  RCV_SHORT_CIRCUIT_MAIN,
  RCV_SHORT_CIRCUIT_PROG,
  RCV_SENSOR_TRIGGERED,
  RCV_SENSOR_RELEASED,
  RCV_S88_STATE,

  RCV_NB_RCV_COMMANDS,
  RCV_NONE      /* special code for NO RCV COMMAND */
}               RCV_CMD_MNEMONIC;

typedef enum
{
  TAG_NONE,
  TAG_RESUME_OPERATIONS,
  TAG_POWER_OFF,
  TAG_EMERGENCY_STOP,
  TAG_LOCO,
  TAG_ACCESSORY,
  TAG_TURNOUT,
  TAG_TURNOUT_QUERY,
  TAG_SIGNAL,
  TAG_DET_SYNC,
  TAG_FUNCTION,
  TAG_WINDOW,
  TAG_INIT_SERVICE_MODE,
  TAG_CV_RD,
  TAG_CV_WR,
  TAG_CSX_DEBUG,
  TAG_CSX_VERSION,
  TAG_POM
}               CMD_TAG_TYPE;

typedef struct S_DCC_CMD_TAG
{
  CMD_TAG_TYPE  nCmdTagType;
  void          *pObjPtr;
  int           nCmdIdx;
  int           nAddress;               /* used in emulation only */
  int           nData;                  /* used in emulation only */
  long          lTimeTickSchedule;
  long          lTimeTickDone;
}               DCC_CMD_TAG;

typedef struct S_DCC_CMD_TAG    *pDCC_CMD_TAG;


typedef struct S_DGI_SERVER_PARAMS
{
  HWND		hMainWindow;
  HANDLE        hComPort;		/* serial port .... */
  int           nComPort;
  void		*pSocket;
  int           nTCPIPPortNum;	         /* ... or TCP IP port */

  int           nIFType;
  int           nBdRateIndex;
  int           nAddress;
  BOOL          bSendBreak;
  BOOL          bEmulationMode;  
  BOOL          bDumpFile; 

  DGI_LANGUAGE  nLanguage;

  int           nSpare[DGI_NB_SRV_PARAM_SPARES];
 
  char          *sExecDirectory;      

}               	DGI_SERVER_PARAMS; /* CDM-RAIL Digital Generic Server */

typedef struct S_DGI_SERVER_PARAMS      *pDGI_SERVER_PARAMS;


typedef struct S_DGI_GENERIC_DATA
{
  DCC_CMD_TAG           *pCmdTag;
  long                  lTimeTick;
  long                  lNbDelayTicks;
  int                   nAddress;
  int                   nAddress2;
  int                   nData[DGI_MAX_DATA_WORDS];
}               DGI_GENERIC_DATA;

typedef struct S_DGI_GENERIC_DATA       *pDGI_GENERIC_DATA;


typedef struct S_DGI_GENERIC_DATA_ITEM
{
  int           nAddress;
  short         hData[4];
}               DGI_GENERIC_DATA_ITEM;

typedef struct S_DGI_GENERIC_RCV_DATA
{
  DCC_CMD_TAG           xCmdTag;
  long                  lTimeTick;
  DCC_COMPLETION_CODE   nCompletionCode;
  RCV_CMD_MNEMONIC      nRdCommand;
  int                   nNbDataItems;
  DGI_GENERIC_DATA_ITEM xDataItem[DGI_MAX_DATA_ITEMS];
}               DGI_GENERIC_RCV_DATA;

typedef struct S_DGI_GENERIC_RCV_DATA   *pDGI_GENERIC_RCV_DATA;
