/*==============================================================*/
/*                                                              */
/*      dccpp_commands.c                                        */
/*      Protocol commands definitions.				*/
/*                                                              */
/*      J.P.PILLOU      13/03/2018                              */
/*==============================================================*/

#include <windows.h>
#include "DDGI_Generic_def.h"
#include "dccpp_def.h"

DCC_COMMAND     xSendCommand[SND_NB_SND_COMMANDS] =     
{
        /* DCCpp INSTRUCTIONS ANS REQUESTS */
        /*====================================*/
        /* header 0xFF means not implemented in DCCpp) */
        /* header 0xFE means fully handled at DLL level (standalone) */
        /* header 0xFD means partially handled at DLL level (local) */
        /* header 0x00 means fully handled by DCCpp */

        /* The response mnemonic convention is:                         */
        /*                                                              */
        /* 1) RCV_ACK: any response from DCC++ other than <X> is OK.    */
        /*      - in that case, CV_ACK is sent back to CDM-Rail.        */
        /*      - on <X>, an error code is sent back to CDM-Rail.       */
        /*                                                              */
        /* 2) RCV_NONE: NO response expected from DCC++.                */
        /*                                                              */
        /* 3) other: response to 

  
        /* SND_RESUME_OPER_REQUEST */
        { 0xFD, '1', 0, 2, { RCV_ACK }, "p1" },

        /* SND_STOP_ALL_OPERS_REQUEST (POWER OFF)*/
        { 0xFD, '0', 0, 2, { RCV_ACK }, "p0" },

        /* SND_STOP_ALL_LOCOS_REQUEST: no command in DCCpp */
        { 0xFD, '\0' , 0, 2, { RCV_ACK }, "" },

        /* SND_STOP_LOCO_V1_V2 */
        { 0xFF },

        /* SND_STOP_LOCO */
        { 0xFD, '\0' , 0, 2, { RCV_ACK }, "" },

        /* SND_ENTER_SERVICE_MODE */
        { 0xFE, '\0' , 0, 0, { RCV_ACK }, "" },             // standalone

        /* SND_DIRECT_CV_READ */
        { 0xFD, 'R' , 3, 30, { RCV_SRVC_MODE_RESP_DIRECT_CV }, "r" },

        /* SND_SERVICE_MODE_EVENT_REQUEST */
        { 0xFE, '\0' , 0, 0, { RCV_SRVC_MODE_RESP_DIRECT_CV }, "" },        // standalone

        /* SND_DIRECT_CV_WRITE */
        { 0xFD, 'W' , 4, 30, { RCV_ACK }, "r" },

        /* SND_EXIT_SERVICE_MODE */
        { 0xFE, '\0' , 0, 0, { RCV_ACK }, "" },             // standalone

        /* SND_CMDSTATION_SOFT_V_REQUEST */
        { 0xFE, '\0' , 0, 0, { RCV_NONE }, "" },              // standalone

        /* SND_CMDSTATION_STATUS_REQUEST */
        { 0xFE, '\0' , 0, 0, { RCV_CMDSTATION_STATUS }, "" }, // standalone

        /* SND_ACCESSORY_INFO_REQUEST */
        { 0xFE, '\0' , 0, 0, { RCV_ACCESSORY_INFO }, "" },    // standalone

        /* SND_DETECTOR_INFO_REQUEST */
        { 0xFD, 'Y' , 0, 0, { RCV_DETECTOR_INFO }, "" },      // standalone    

        /* SND_ACCESSORY_OPER */
        { 0xFD, 'T', 2, 2, { RCV_ACK }, "H" }, 

        /* SND_LOCO_INFO_REQUEST */
        { 0xFE, '\0' , 0, 0, { RCV_NORMAL_LOCO_INFO }, "" },  // standalone

        /* SND_LOCO_F1F8_STATUS_REQUEST */
        { 0xFF },
        
        /* SND_LOCO_OPER_V2 */
        { 0xFF },
        
        /* SND_LOCO_SPEED_14 */
        { 0xFF },
        
        /* SND_LOCO_SPEED_27 */
        { 0xFF },
        
        /* SND_LOCO_SPEED_28 */
        { 0xFF },
        
        /* SND_LOCO_SPEED_128 */
        { 0xFD, 't', 4, 2, { RCV_ACK }, "T" }, 
        
        /* SND_FUNCTION_GROUP_1 */
        { 0xFD, 'f', 2, 2, { RCV_NONE } },
        
        /* SND_FUNCTION_GROUP_2 */
        { 0xFD, 'f', 2, 2, { RCV_NONE } },
        
        /* SND_FUNCTION_GROUP_3 */
        { 0xFD, 'f', 2, 2, { RCV_NONE } },
       
        /* SND_DEFINE_TURNOUT */
        { 0xFD, 'T', 3, 2, { RCV_ACK }, "O" },
       
        /* SND_STORE_IN_EEPROM */
        { 0xFD, 'E', 0, 2, { RCV_ACK }, "e" },
       
        /* SND_INIT_S88 */
        { 0xFD, 'Y', 2, 2, { RCV_ACK }, "y" },

        /* SND_S88_UPDATE_REQUEST */
        { 0xFD, 'Y' , 0, 2, { RCV_S88_STATE }, "y" },  

       
};
                      
/* receive codes not used with DCC++ */                                  
DCC_COMMAND     xRecvCommand[RCV_NB_RCV_COMMANDS];    
