/*======================================================================*/
/*                                                                      */
/*      dccpp_def.h                                                     */
/*      Header file of dccpp_def.c ( Main module of DLL DDGI_DCCpp ).   */
/*======================================================================*/
/*      J.P.PILLOU      13/03/2018                                      */
/*======================================================================*/

#define DLL_NAME        "DCC++"
#define DLL_VERSION     "V6.01b (2018/03/29)"
      
                                                                     
#define MAX_COM_PORT_NUM        50
#define MAX_COM_PORT_NAME_LEN   10
#define MAX_STATION_TYPE_LEN    32
#define MAX_VERSION_STRING_LEN  32
#define MAX_STR_LEN             128
#define MAX_MSG_LEN             128
#define MAX_LONG_MSG_LEN        1024
#define MAX_SND_MSG             256
#define MAX_RCV_MSG             64
#define MAX_ERR_MSG             64
#define RCV_FIFO_SIZE           1024
#define RCV_DUMP_SIZE           256
#define MAX_RD_NB_ITERS         3       /* number of iterations for read */

#define MAX_DCCPP_CMD_LEN       256
#define MAX_NB_CMD_DATA	        8
#define MAX_NB_CMD_BYTES	512
#define MAX_NB_DCC_DATA_WORDS   3

#define MAX_NB_DETECTORS        32


#define MAX_DBG_BUFFER_NB_LINES 128 
//#define MAX_DBG_FLUSH_NB_LINES          96 unused
#define MAX_DBG_DISPLAY_NB_LINES        32

#define DATA_COM_TIME_OUT       ((DWORD) 200)
#define T_100_MS                ((DWORD) 100)

/* specific time out for initial access to CS, ... */
#define INIT_PHASE_TIME_OUT     ((DWORD) 1000)

#define DUMP_FILE_NAME          "DCCpp.txt"
#define DEBUG_FILE_NAME         "DBG_DCCpp.txt"

#define DCC_MAX_SHORT_ADDRESS           99
#define DCC_ADDRESS_OFFSET              (unsigned short) 0xC000
#define DCC_MAX_ACCESSORY_ADDRESS       (unsigned short) 2048
#define DCC_MAX_TURNOUT_ADDRESS         256
#define DCC_MIN_FEEDBACK_ADDRESS        513
#define MAX_NB_RESPONSES                6       
#define RCV_FEEDBACK_INDEX              10

#define LOCO_DIR_BIT_MASK               ((unsigned char) 32 )            
#define LOCO_F0_BIT_MASK                ((unsigned char) 16 )
#define TURNOUT_STATE_BIT_MASK          ((unsigned char) 4 )

#define DUMMY_LOCO_ADDRESS              60
#define REFRESH_INTERVAL                2

#define DCCPP_BAUD_RATE                 CBR_115200 

#define MAX_NB_DCCpp_LOC_REGS           12               

typedef enum
{
  SND_RESUME_OPER_REQUEST,
  SND_STOP_ALL_OPERS_REQUEST,
  SND_STOP_ALL_LOCOS_REQUEST,
  SND_STOP_LOCO_V1_V2,
  SND_STOP_LOCO,
  SND_ENTER_SERVICE_MODE,
  SND_DIRECT_CV_READ,
  SND_SERVICE_MODE_EVENT_REQUEST,
  SND_DIRECT_CV_WRITE,
  SND_EXIT_SERVICE_MODE,
  SND_CMDSTATION_SOFT_V_REQUEST,
  SND_CMDSTATION_STATUS_REQUEST,
  SND_ACCESSORY_INFO_REQUEST,
  SND_DETECTOR_INFO_REQUEST,
  SND_ACCESSORY_OPER,
  SND_LOCO_INFO_REQUEST,
  SND_LOCO_F1F8_STATUS_REQUEST,
  SND_LOCO_OPER_V2,
  SND_LOCO_SPEED_14,
  SND_LOCO_SPEED_27,
  SND_LOCO_SPEED_28,
  SND_LOCO_SPEED_128,
  SND_FUNCTION_GROUP_1,
  SND_FUNCTION_GROUP_2,
  SND_FUNCTION_GROUP_3,
  
  // specific DCCpp (local commands)
  SND_DEFINE_TURNOUT,
  SND_STORE_IN_EEPROM,
  SND_INIT_S88,
  SND_S88_UPDATE_REQUEST,
  SND_NB_SND_COMMANDS
}               SND_CMD_MNEMONIC;

typedef enum
{
  DCC_SEQ_NONE,
  DCC_SEQ_SCHEDULED,
  DCC_SEQ_INIT,
  DCC_SEQ_SENT,
  DCC_SEQ_DONE,
  DCC_SEQ_LOCAL_DONE
}               DCC_SEQ_STATE;

typedef enum
{
  RD_RESPONSE,
  RD_FEEDBACK,
  RD_PURGE
}               RD_PHASE;

typedef enum
{
  PRI_0,
  PRI_1,
  PRI_2,
  PRI_3
}               CMD_PRIORITY;



typedef enum
{
  CMD_DATA_NONE,
  CMD_DATA_CMD,
  CMD_DATA_DEC,
  CMD_DATA_HEX_BYTE,
  CMD_DATA_HEX_BYTE_PREFIX,
  CMD_DATA_HEX_WORD,
  CMD_DATA_HEX_WORD_PREFIX
}               CMD_DATA_TYPE;              

typedef enum
{
  RD_ERR_NONE,
  RD_ERR_NO_RESPONSE,
  RD_ERR_NO_PROMPT,
  RD_ERR_OVERLOAD,
  RD_ERR_CV_DIRECT_ERROR,
  RD_ERR_CV_PAGED_ERROR,
  RD_ERR_ARG_FORMAT_ERROR,
  RD_ERR_UNRECOGNIZED_COMMAND,
  RD_ERR_ADDRESS_TOO_LARGE,
  RD_ERR_SPEED_STEP_TOO_LARGE,
  RD_ERR_VALUE_TOO_LARGE,
  RD_ERR_BAD_BOOT_COMMAND,
  RD_ERR_ALREADY_PROGRAMMING,
  RD_ERR_UNDETERMINED_ERROR,
  RD_ERR_UNEXPECTED_RESPONSE,
  RD_ERR_BAD_NUMERIC_RESPONSE,
 }               RD_ERROR;

typedef enum
{
  CMD_NORMAL,
  CMD_STANDALONE,
  CMD_LOCAL,
  CMD_FEEDBACK,
  CMD_LOCAL_FEEDBACK,
  CMD_REFRESH
}               CMD_TYPE;





typedef struct S_SERVER_DESC
{
  int                   nComPortNum;
  int                   nComPortBaudRate;
  /* DCC_PROT_TYPE         nProtType; */
  unsigned char         sStationType[MAX_STATION_TYPE_LEN];
  unsigned char         sVersionString[MAX_VERSION_STRING_LEN];
  RCV_CMD_MNEMONIC      nLastRdCommand;    /* in case of protocol error */
  RD_ERROR              nLastRdErrorCode;  /* in case of comm error */

  BOOL                  bEmergencyStop;
  BOOL                  bPowerOff;
                                        /* new 12/2007 */
  long                  lCurTimeTick;
}                       SERVER_DESC;

typedef struct S_DCC_COMMAND    *pDCC_COMMAND;

typedef struct S_CMD_DATA
{
  CMD_DATA_TYPE	        nType;
  unsigned char	        cValue;
  unsigned int          iValue;
  unsigned char         cHiVOffset;
}                       CMD_DATA;


typedef struct S_DCC_COMMAND
{
  unsigned char         cHeaderCode;    // implented or NOT,...
  unsigned char         cCommandCode;   // DCC++ command symbol
  int                   nNbArgs;        // nb od DCC++ arguments
  int                   nLatency;       // max allowed delay for response (in 100ms units)
  RCV_CMD_MNEMONIC      nResponse[MAX_NB_RESPONSES];
  char                  sExpectedRespHeader[3];
  int                   xDataByte[MAX_NB_CMD_DATA]; 
}		DCC_COMMAND;

typedef struct S_DCC_COMMAND_CTXT       *pDCC_COMMAND_CTXT;
typedef struct S_DCC_CMD_CTXT_ENV       *pDCC_CMD_CTXT_ENV;

typedef struct S_DCC_COMMAND_CTXT
{
  DCC_COMMAND           xSndCommand;
  DCC_COMMAND           xRcvCommand;
  long                  lIndex;
  DCC_SEQ_STATE         nSndSeqState;
  BOOL                  bDisplay;
  CMD_TYPE              nCommandType;
  DCC_COMPLETION_CODE   nCompletionCode;
  RD_ERROR              nRdErrorCode;
  SND_CMD_MNEMONIC      nSndCommand;
  RCV_CMD_MNEMONIC      nRdCommand;
  DCC_CMD_TAG           xCmdTag;
  int                   nNumber;
  CMD_PRIORITY          nPriority;
  pDCC_COMMAND_CTXT     pPrev;
  pDCC_COMMAND_CTXT     pNext;
}                       DCC_COMMAND_CTXT;

typedef struct S_DCC_CMD_CTXT_ENV
{
  pDCC_COMMAND_CTXT     pCmdCtxtList;
  pDCC_COMMAND_CTXT     pCmdWrPtr;   
  pDCC_COMMAND_CTXT     pCmdRdPtr;
  int                   nNbCmd;
  int                   nNbPendingCmd;
}                       DCC_CMD_CTXT_ENV;


typedef struct S_DCC_CONTEXT
{
  SERVER_DESC           xServerDesc;
  HANDLE                hComPort;
  HANDLE                hIFThread;
  DWORD                 lIFThreadID;
  CRITICAL_SECTION      xCriticalSection;       /* spinlock object */

  long                  lTotalNbCmdSent;
  long                  lTotalNbRefreshCmdSent;
  long                  lTotalNbLocalCmdSent;
  long                  lTotalNbUserCmdSent;
  long                  lTotalNbErrors;
  long                  lTotalNbMsgReceived;
  long                  lTotalNbCmdBusyMsg;

  long                  lTotalDeltaTime;

  long                  lNbSndOverflows;
  long                  lNbErrOverflows;
  long                  lNbRcvOverflows;

  DCC_COMMAND_CTXT      xCurCmdResponse;        /* used on non ACK response */

  DCC_CMD_CTXT_ENV      xSndCmd;
  DCC_CMD_CTXT_ENV      xError;
  DCC_CMD_CTXT_ENV      xRcvMsg;

  pDCC_COMMAND_CTXT     pFreeCmdCtxtList;
  int                   nNbFreeCmdCtxt;

  /*
  pDCC_COMMAND_CTXT     pSndCmdCtxtList;
  pDCC_COMMAND_CTXT     pSndCmdWrPtr;   
  pDCC_COMMAND_CTXT     pSndCmdRdPtr;
  int                   nNbSndCmd;
  int                   nNbPendingSndCmd;
  */

  /*
  DCC_COMMAND_CTXT      xSndMsgFifo[MAX_SND_MSG];
  int                   nSndMsgWrPtr;
  int                   nSndMsgRdPtr;
  int                   nSndMsgRd2Ptr;
  int                   nNbPendingSndMsg;
  */

  /*
  unsigned char         xRcvByteFifo[RCV_FIFO_SIZE];
  int                   nRcvByteWrPtr;
  int                   nRcvByteRdPtr;
  */

  DCC_COMMAND           xRcvMsgFifo[MAX_RCV_MSG];
  int                   nRcvMsgWrPtr;
  int                   nRcvMsgRdPtr;
}                       DCC_CONTEXT;

typedef struct S_DBG_BUFFER
{
  long                  lTimeTick;
  BOOL                  bWrite;
  BOOL                  bBroadcast;     
  BOOL                  bResponsePatched;  
  CMD_TYPE              nCommandType;
  int                   nNbBytes;
  int                   nFeedbackGenState;
  int                   nSubResponseIndex;
  unsigned char         cBuffer[MAX_NB_CMD_DATA + 1];
}                       DBG_BUFFER;

typedef struct S_DETECTOR_STATE
{
  long                  lTimeTick;
  unsigned char         cDetectorStateFromBroadcast;
  unsigned char         cDetectorState;
  unsigned char         cSpare[2];
}                       DETECTOR_STATE;




