/*======================================================================*/
/*                                                                      */
/*      dccpp_specific.h                                                */
/*      Header file of dccpp_specific.c                                 */
/*      Module containing miscellaneous utilities related to DCC++      */
/*======================================================================*/
/*      J.P.PILLOU      13/03/2018                                      */
/*======================================================================*/
/*======================================================================*/

/*======================================================================*/
/*                              TYPE DEFINITIONS                        */
/*======================================================================*/

#define MAX_NB_ACCESSORIES      2048

#define ACC_THROWN_BIT          0
#define ACC_UNTHROWN_BIT        1
#define ACC_DEFINED_BIT         7

#define ACC_THROWN_MASK         ((unsigned char) 1)
#define ACC_UNTHROWN_MASK       ((unsigned char) 2)
#define ACC_DEFINED_MASK        ((unsigned char) 128)

typedef struct S_LOCO_STRUCT   *pLOCO_STRUCT;

typedef struct S_LOCO_STRUCT
{
  int                   nAddress;
  int                   nSpeed;
  BOOL                  bForward;
  int                   nF0F12;
  int                   nF13F28;
  BOOL                  bEmergencyStop;
  SPEED_STEP_MODE       nMode;  // unused in DCC++ - default 128 steps
  int                   nRegister;
  long                  lLastTimeTick;
  pLOCO_STRUCT          pPrev;
  pLOCO_STRUCT          pNext;
}                       LOCO_STRUCT;



/*======================================================================*/
/*                          FUNCTION DECLARATIONS                       */
/*======================================================================*/

/* LOCOS UTILITIES */

BOOL    DCPS_EmergencyStopAllLocos( DCC_CMD_TAG         *pCmdTag );

BOOL    DCPS_RestartAllLocos( void );


unsigned char
        DCPS_SetFxByte(         unsigned int            nFuncValue, 
                                unsigned int            nFuncMask );

int     DCPS_UpdateLocoParameters( int                  nAddress, 
                                int                     nSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode, 
                                BOOL                    bForward,
                                unsigned int            nF0F12,
                                unsigned int            nF0F12Mask );

void    DCPS_UpdateLocoSpeed(   int                     nAddress, 
                                int                     nSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode );

void    DCPS_UpdateLocoFx(      int                     nAddress, 
                                unsigned int            nF0F12,
                                unsigned int            nF0F12Mask );

void    DCPS_SetEmergencyStop(  int                     nAddress );

BOOL    DCPS_GetLocoInfo(       pDGI_GENERIC_RCV_DATA   pGenericRcvData, 
                                int                     nAddress );

BOOL    DCPS_GetLocoSpeedMode(  int                     nAddress, 
                                SPEED_STEP_MODE         *pSpeedStepMode );


/* ACCESSORY FUNCTIONS */

BOOL    DCPS_AccessoryDefined(  int                     nAddress );

void    DCPS_SetAccessoryState( int                     nAddress, 
                                BOOL                    bThrown );

BOOL    DCPS_AccessoryState(    int                     nAddress );

BOOL    DCPS_GetAccessoryInfo(  pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                int                     nAddress );


/* LIST UTILITIES */


pLOCO_STRUCT
        DCPS_AddLoco(           int                     nAddress );

pLOCO_STRUCT
        DCPS_FindLoco(          int                     nAddress );


/* S88 UTILITIES */

void    DCPS_InitS88(           void );

void    DCPS_S88UpdateRequest(  void );

BOOL    DCPS_ExtractS88State(   unsigned char           *xRcvByte );

BOOL    DCPS_S88StateChanged(   void );


