/*======================================================================*/
/*                                                                      */
/*      dccpp_util.c                                                    */
/*      This module contains miscellaneous utilities related to DCC++   */
/*======================================================================*/
/*      J.P.PILLOU      18/03/2018                                      */
/*======================================================================*/
/*======================================================================*/

#include <windows.h>
#include <stdio.h>
#include "DDGI_Generic_def.h"
#include "dccpp_def.h"
#include "dccpp_extern.h"
#include "dccpp_util.h"

/*======================================================================*/
/*                         FUNCTION DECLARATIONS                        */
/*======================================================================*/

char    *DCPU_CompletionCodeStr(
                                DCC_COMPLETION_CODE     nCompletionCode );
                                
char    *DCPU_RcvCmdMnemonicStr(
                                RCV_CMD_MNEMONIC        nRdCommand );

char    *DCPU_SndCmdMnemonicStr(
                                SND_CMD_MNEMONIC        nSndCommand );

char    *DCPU_RdErrorStr(       RD_ERROR            nRdError );

BOOL    DCPU_ReadFile(          HANDLE                  hCom, 
                                unsigned char           *pBuffer,
                                LPDWORD                 pNbBytesRead,
                                RD_PHASE                nPhase,
                                BOOL                    bStoreInFifo );

BOOL    DCPU_WriteFile(         HANDLE                  hCom, 
                                unsigned char           *pBuffer, 
                                DWORD                   nNbBytes, 
                                LPDWORD                 pNbBytesRead,
                                BOOL                    bStoreInFifo );

void    DCPU_CopyBufferToFifo(  char                    *pOrgPtr,
                                int                     nNbBytes,
                                char                    cSeparator,
                                BOOL                    bWrite,
                                BOOL                    bNewCmd,
                                CMD_TYPE                nCmdType );
                                         
void    DCPU_RemoveLastCmdFifoEntry( void );

void    DCPU_UpdateLastCmdFifoEntry( DCC_COMMAND_CTXT   *pSndCommandCtxt );

void    DCPU_SetLastCmdFifoEntry( CMD_TYPE              nCmdType );

void    DCPU_DumpDbgFifo(       HWND                    hWnd );

void    DCPU_DumpDbgFifoToFile( FILE                    *pFile );

void    DCPU_DumpCmdFifo(       HWND                    hWnd,
                                FILE                    *pFile,
                                BOOL                    bError );

void    DCPU_DumpCmdFifoToFile( FILE                    *pFile );

void    DCPU_PrintMessageToFile( FILE                   *pFile, 
                                char                    sMessage[] );

void    DCPU_ClearCmdTag(       pDCC_CMD_TAG            pCmdTag );

void    DCPU_CopySndCommand(    SND_CMD_MNEMONIC        nCommand, 
                                int                     nNbArgs,
                                int                     *pDataArray,
                                DCC_COMMAND             *pCommand );

BOOL    DCPU_IsExpectedResponse( RCV_CMD_MNEMONIC       nRdCommand, 
                                DCC_COMMAND             *pRcvCommand,
                                DCC_COMMAND             *pSndCommand );

int     DCPU_RcvAccessoryAddress( CMD_DATA              xDataByte[] );

int     DCPU_SndAccessoryAddress( CMD_DATA              xDataByte[] );


/* DCC_COMMAND_CTXT HANDLING */

pDCC_COMMAND_CTXT 
        DCPU_CreateCommandCtxt( void );

void    DCPU_ClearCommandCtxt(  pDCC_COMMAND_CTXT       pCmdCtxt );

void    DCPU_MapResponseToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                pDCC_COMMAND_CTXT       pCmdCtxt,
                                RCV_CMD_MNEMONIC        *pCommand );

void    DCPU_MapAccessoryInfoToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData );

void    DCPU_MapDetectorInfoToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData );

void    DCPU_SwapNibbleBits(            
                                unsigned char           cValue, 
                                unsigned short          *pValue );


void    DCPU_MapLocoInfoToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData );

void    DCPU_MapCVInfoToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                DCC_COMMAND_CTXT        *pCmdCtxt,
                                RCV_CMD_MNEMONIC        *pCommand );

void    DCPU_MapCSStatusToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                DCC_COMMAND_CTXT        *pCmdCtxt );

void    DCPU_MapMessageToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                pDCC_COMMAND_CTXT       pCmdCtxt,
                                RCV_CMD_MNEMONIC        *pCommand );

void    DCPU_MapFeedbackBroadcastToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData );


void    DCPU_CopyCommanCtxt(    pDCC_COMMAND_CTXT       pDstCmdCtxt, 
                                pDCC_COMMAND_CTXT       pSrcCmdCtxt );

void    DCPU_ClearCommand(      pDCC_COMMAND            pCommand );

pDCC_COMMAND_CTXT
        DCPU_NewCommandCtxt(    pDCC_CMD_CTXT_ENV       pCmdEnv,
                                CMD_PRIORITY            nPriority );

void    DCPU_DetachCommandCtxt( pDCC_CMD_CTXT_ENV       pCmdEnv );

void    DCPU_FreeCommandCtxtList( pDCC_CMD_CTXT_ENV     pCmdEnv );


/* DCC FORMAT CONVERSION UTILITIES */

int     DCPU_F0F12Byte2Int(     unsigned char           cFuncA, 
                                unsigned char           cFuncB );

unsigned char   
        DCPU_SpeedInt2Byte(     int                     nSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode );

int     DCPU_SpeedByte2Int(     unsigned char           cSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode,
                                BOOL                    bReverse );

int     DCPU_AddressByte2Int(   unsigned char           cADH, 
                                unsigned char           cADL );

/* DATA COM UTILITIES */

void    DCPU_SetTimeOut(        HANDLE                  hCom, 
                                DWORD                   nTimeOut );

/* STRING UTILITIES */

BOOL    DCPU_SkipBlanks(        unsigned char           xBuffer[], 
                                int                     *pIndex );

BOOL    DCPU_GetNextCharIndex(  unsigned char           xBuffer[], 
                                unsigned char           cExpectedChar[],
                                int                     *pIndex );

BOOL    DCPU_IsDecimal(         unsigned char           cChar );

BOOL    DCPU_IsHexaDecimal(     unsigned char           cChar );

BOOL    DCPU_ExtractDigits(     unsigned char           xRcvBuffer[], 
                                int                     *pIndex, 
                                int                     nNbDigits, 
                                unsigned char           xDigitArray[] );

unsigned char
        DCPU_SetByte(           unsigned char           cHigh, 
                                unsigned char           cLow );

BOOL    DCPU_Ascii2Hexa(        unsigned char           *pBuffer, 
                                int                     nNbChars, 
                                unsigned int            *pValue );


unsigned char *
        DCPU_DisplayVersion(    unsigned char           xByte[], 
                                int                     nNbDigits, 
                                int                     nAfterDot,
                                unsigned char           xBuffer[] );

BOOL    DCPU_IsKnownErrorCode(  unsigned char           cByte, 
                                int                     *pErrorIdx );

BOOL    DCPU_CommandStationBusy( RD_ERROR           nRdError );

void    DCPU_PrintDebugMessage( char                    *sDbgMsg );
                                
/*======================================================================*/
/*                             STATIC VARIABLES                         */
/*======================================================================*/

char            cRcvFifo[RCV_FIFO_SIZE];
int             nRcvFifoWrPtr = 0;

DBG_BUFFER      xCmdFifo[MAX_DBG_BUFFER_NB_LINES];
int             nCmdFifoWrPtr = 0; 

int             nTotalCmdNb = 0;
int             nLastFlushCmdIndex = 0; 

char            *sCommandPrefix[][2] =
                { { "SEND", "RESP" },   /* normal instruction */
                  { "SSND", "SRSP" },   /* standalone */
                  { "LSND", "LRSP" },   /* local */
                  { "XXXX", "FBCK" },   /* feedback: none for SPROG */
                  { "XXXX", "LFBK" } };   /* local feedback */

/*======================================================================*/
/*                              FUNCTION DEFINITIONS                    */
/*======================================================================*/

/*======================================================================*/
/*      DCPU_CompletionCodeStr                                          */
/*                                                                      */
/*      RETURN: string describing completion code.                      */

char            *DCPU_CompletionCodeStr(
                                DCC_COMPLETION_CODE     nCompletionCode )
{
  char                  *pCompCodeString = "NOT FOUND";

  switch( nCompletionCode )
  {
  case DGI_COMP_NONE:
    pCompCodeString = "C_NONE";
    break;

  case DGI_COMP_OK:
    pCompCodeString = "C_OK";
    break;

  case DGI_COMP_OK_RD:
    pCompCodeString = "C_OK_RD";
    break;

  case DGI_COMP_READ_ERR:
    pCompCodeString = "C_READ_ERR";
    break;

  case DGI_COMP_PROT_ERR:
    pCompCodeString = "C_PROT_ERR";
    break;

  case DGI_COMP_PROG_ERR:
    pCompCodeString = "C_PROG_ERR";
    break;

  case DGI_COMP_MUDH_ERR:
    pCompCodeString = "C_PROT_ERR";
    break;

  case DGI_COMP_OPER_ERR:
    pCompCodeString = "C_PROT_ERR";
    break;

  default:
    pCompCodeString = "C_NOT_FOUND";
  }

  return pCompCodeString;
}
                                
/*======================================================================*/
/*      DCPU_RcvCmdMnemonicStr                                          */
/*                                                                      */
/*      RETURN: string describing read command                          */

char            *DCPU_RcvCmdMnemonicStr(
                                RCV_CMD_MNEMONIC        nRdCommand )
{
  char                  *pRdCommandString = "NOT FOUND";

  switch( nRdCommand )
  {
  case RCV_PC_IF_ERROR:
    pRdCommandString = "R_PC_IF";
    break;

  case RCV_IF_CS_ERROR:
    pRdCommandString = "R_IF_CS";
    break;

  case RCV_UNKNOWN_COMM_ERROR:
    pRdCommandString = "R_U_CMD";
    break;

  case RCV_ACK:
    pRdCommandString = "R_ACK";
    break;

  case RCV_NO_CS_TIME_SLOT:
    pRdCommandString = "R_NO_TS";
    break;

  case RCV_IF_BUFFER_OVERFLOW:
    pRdCommandString = "R_IF_OVF";
    break;

  case RCV_NORMAL_OPER_RESUMED:
    pRdCommandString = "RESUME";
    break;

  case RCV_TRACK_POWER_OFF:
    pRdCommandString = "P_OFF";
    break;

  case RCV_EMERGENCY_STOP:
    pRdCommandString = "EM_STOP";
    break;

  case RCV_SERVICE_MODE_ENTRY:
    pRdCommandString = "SRVC_ENTRY";
    break;

  case RCV_FEEDBACK_BROADCAST:
    pRdCommandString = "FBCK_BRDCST";
    break;

  case RCV_PROG_SHORT_CIRCUIT:
    pRdCommandString = "R_PRG_SCKT";
    break;

  case RCV_PROG_DATA_BYTE_NOT_FOUND:
    pRdCommandString = "R_PRG_NO_D";
    break;

  case RCV_PROG_CMDSTATION_BUSY:
    pRdCommandString = "R_S_BUSY";
    break;

  case RCV_PROG_CMDSTATION_READY:
    pRdCommandString = "R_S_READY";
    break;

  case RCV_SRVC_MODE_RESP_REG_PAGED:
    pRdCommandString = "R_PAGE_CV";
    break;

  case RCV_SRVC_MODE_RESP_DIRECT_CV:
    pRdCommandString = "R_DIR_CV";
    break;

  case RCV_CMDSTATION_SOFT_V_ANSWER_XBUS:
    pRdCommandString = "XXX";
    break;

  case RCV_CMDSTATION_SOFT_V_ANSWER_XPNET:
    pRdCommandString = "XXX";
    break;

  case RCV_CMDSTATION_STATUS:
    pRdCommandString = "R_STATION_STATUS";
    break;

  case RCV_TRANSFER_ERRORS:
    pRdCommandString = "R_XFER_ERR";
    break;

  case RCV_CMDSTATION_BUSY:
    pRdCommandString = "R_BUSY";
    break;

  case RCV_UNSUPPORTED_COMMAND:
    pRdCommandString = "R_UNSUP_CMD";
    break;

  case RCV_ACCESSORY_INFO:
    pRdCommandString = "R_ACC_INFO";
    break;

  case RCV_DETECTOR_INFO:
    pRdCommandString = "R_DET_INFO";
    break;

  case RCV_NORMAL_LOCO_INFO:
    pRdCommandString = "R_LOCO_INFO";
    break;

  case RCV_MULTI_UNIT_LOCO_INFO:
    pRdCommandString = "R_XXX";
    break;

  case RCV_MULTI_UNIT_AD_LOCO_INFO:
    pRdCommandString = "R_XXX";
    break;

  case RCV_DOUBLE_HEADER_LOCO_INFO:
    pRdCommandString = "R_XXX";
    break;

  case RCV_LOCO_OPER_BY_ANOTHER_DEVICE:
    pRdCommandString = "R_LOCO_CHG";
    break;

  case RCV_FUNCTION_STATUS:
    pRdCommandString = "R_FUNC_ST";
    break;

  case RCV_LOCO_AD_RETRIEVAL:
    pRdCommandString = "R_LOCO_AD_RET";
    break;

  case RCV_DOUBLE_HEADER_ERROR:
    pRdCommandString = "R_DH_ERROR";
    break;

  case RCV_MU_DH_ERROR:
    pRdCommandString = "R_MU_DH_ERROR";
    break;

  case RCV_NB_RCV_COMMANDS:
    pRdCommandString = "XXX";
    break;

  case RCV_NONE:
    pRdCommandString = "RCV_NONE";
    break;

  default:
    pRdCommandString = "R_NOT_FOUND";

  }

  return pRdCommandString;
}

/*======================================================================*/
/*      DCPU_SndCmdMnemonicStr                                          */
/*                                                                      */
/*      RETURN: string describing send command                          */

char            *DCPU_SndCmdMnemonicStr(
                                SND_CMD_MNEMONIC        nSndCommand )
{
  char                  *pSndCommandString = "NOT FOUND";

  switch( nSndCommand )
  {
  case SND_RESUME_OPER_REQUEST:
    pSndCommandString = "SND_RESUME_OPER_REQUEST";
    break;
    
  case SND_STOP_ALL_OPERS_REQUEST:
    pSndCommandString = "SND_STOP_ALL_OPERS_REQUEST";
    break;
    
  case SND_STOP_ALL_LOCOS_REQUEST:
    pSndCommandString = "SND_STOP_ALL_LOCOS_REQUEST";
    break;
    
  case SND_ENTER_SERVICE_MODE:
    pSndCommandString = "SND_ENTER_SERVICE_MODE";
    break;
    
  case SND_DIRECT_CV_READ:
    pSndCommandString = "SND_DIRECT_CV_READ";
    break;
    
  case SND_SERVICE_MODE_EVENT_REQUEST:
    pSndCommandString = "SND_SERVICE_MODE_EVENT_REQUEST";
    break;
    
  case SND_DIRECT_CV_WRITE:
    pSndCommandString = "SND_DIRECT_CV_WRITE";
    break;
    
  case SND_EXIT_SERVICE_MODE:
    pSndCommandString = "SND_EXIT_SERVICE_MODE";
    break;
    
  case SND_ACCESSORY_INFO_REQUEST:
    pSndCommandString = "SND_ACCESSORY_INFO_REQUEST";
    break;
    
  case SND_ACCESSORY_OPER:
    pSndCommandString = "SND_ACCESSORY_OPER";
    break;
    
  case SND_LOCO_INFO_REQUEST:
    pSndCommandString = "SND_LOCO_INFO_REQUEST";
    break;
    
  default:
    pSndCommandString = "NOT FOUND";
    break;
  }

  return pSndCommandString;
}

/*======================================================================*/
/*      DCPU_RdErrorStr                                                 */
/*                                                                      */
/*      RETURN: string describing error code returned after read oper.  */

char            *DCPU_RdErrorStr( RD_ERROR      nRdError )
{
  char                  *pErrorString = "NOT FOUND";

  switch( nRdError )
  {
  case RD_ERR_NONE:
    pErrorString = "E_NO_ERR";
    break;

  case RD_ERR_NO_RESPONSE:
    pErrorString = "E_NO_RESP";
    break;

  case RD_ERR_NO_PROMPT:
    pErrorString = "E_NO_PROMPT";
    break;

  case RD_ERR_OVERLOAD:
    pErrorString = "E_OVRLD";
    break;

  case RD_ERR_CV_DIRECT_ERROR:
    pErrorString = "E_D_CV";
    break;

  case RD_ERR_CV_PAGED_ERROR:
    pErrorString = "E_P_CV";
    break;

  case RD_ERR_ARG_FORMAT_ERROR:
    pErrorString = "E_ARG_FMT";
    break;

  case RD_ERR_UNRECOGNIZED_COMMAND:
    pErrorString = "E_UNRC_CMD";
    break;

  case RD_ERR_ADDRESS_TOO_LARGE:
    pErrorString = "E_AD_2_LRG";
    break;

  case RD_ERR_SPEED_STEP_TOO_LARGE:
    pErrorString = "E_SS_2_LRG";
    break;

  case RD_ERR_VALUE_TOO_LARGE:
    pErrorString = "E_VAL_2_LRG";
    break;

  case RD_ERR_BAD_BOOT_COMMAND:
    pErrorString = "E_BOOT_CMD";
    break;

  case RD_ERR_ALREADY_PROGRAMMING:
    pErrorString = "E_ALRDY_PRG";
    break;

  case RD_ERR_UNDETERMINED_ERROR:
    pErrorString = "E_UNDET";
    break;

  case RD_ERR_UNEXPECTED_RESPONSE:
    pErrorString = "E_UNXP_RESP";
    break;

  case RD_ERR_BAD_NUMERIC_RESPONSE:
    pErrorString = "E_BAD_NUM";
    break;

  default:
    pErrorString = "unknown error code";
  }

  return pErrorString;
}

/*======================================================================*/
/*      DCPU_WriteFile                                                  */
/*      This is a wrapper for WriteFile (base read function), but with  */
/*      copy to debug fifo, IF flag bStoreInFifo is TRUE.               */

BOOL
DCPU_WriteFile( HANDLE hCom, unsigned char *pBuffer, 
                DWORD nNbBytes, LPDWORD pNbBytesWritten, BOOL bStoreInFifo )
{
  BOOL          bOK;

  bOK = WriteFile( hCom, pBuffer, nNbBytes, pNbBytesWritten, NULL );
  if ( bOK && bStoreInFifo )
    DCPU_CopyBufferToFifo( (char *) pBuffer,
                           (int) *pNbBytesWritten,
                           (char) 0xAA,
                           TRUE /* bWrite */,
                           TRUE /* bNewCmd */,
                           CMD_NORMAL );
  
  return bOK;
}

/*======================================================================*/
/*      DCPU_ReadFile                                                   */
/*      This is a wrapper for ReadFile (base read function), but with   */
/*      copy to debug fifo.                                             */
/*                                                                      */
/*      When invoked for reading an expected response                   */
/*      (phase RD_RESPONSE):                                            */
/*      Since the number of chars to be read is NOT known, Readfile     */
/*      is repeatedly invoked for on single char read, until the        */
/*      closing bracket is read.                                        */
/*                                                                      */
/*      When invoked for looking for a feedback message                 */
/*      (phase RD_FEEDBACK):                                            */
/*      The function
/*      Since the number of chars to be read is NOT known, Readfile     */
/*      is repeatedly invoked for on single char read, until the        */
/*      closing bracket is read.                                        */
/*                                                                      */
/*      RETURN: TRUE if prompt sign found.                              */

BOOL
DCPU_ReadFile(  HANDLE hCom, unsigned char *pBuffer,
                LPDWORD pNbBytesRead, RD_PHASE nPhase, BOOL bStoreToFifo )
{
  BOOL          bOK;
  int           nIter = 0;
  BOOL          bDone = FALSE;
  DWORD         nNbBytesRead;
  DWORD         nTotalNbBytesRead;
  int           i;
  unsigned char cSeparator;

  if ( nPhase == RD_RESPONSE )          // expected response
  {
    nTotalNbBytesRead = 0;
    bDone = FALSE;

    while ( !bDone )
    {
      nNbBytesRead = 0;
      bOK = ReadFile( hCom, &pBuffer[nTotalNbBytesRead], 1, &nNbBytesRead, NULL );
      if ( bOK && nNbBytesRead != 0 )
      {
        nTotalNbBytesRead++;    
        if ( pBuffer[nTotalNbBytesRead - 1] == '>' )      // last char
          bDone = TRUE;
      }
      else
        bDone = TRUE;
    }
  }

  else  // RD_FEEDBACK
  {
    // First read with 0 latency 
    DCPU_SetTimeOut( hCom, 0 );

    nTotalNbBytesRead = 0;
    bOK = ReadFile( hCom, &pBuffer[nTotalNbBytesRead], 1, &nNbBytesRead, NULL );
    if ( bOK && nNbBytesRead != 0 )     // header '<' read
    {
      nTotalNbBytesRead++;
      // for fetching message remainder, restore time-out , which has set to 0
      DCPU_SetTimeOut( hCom, DATA_COM_TIME_OUT );

      bDone = FALSE;

      while ( !bDone )
      {
        nNbBytesRead = 0;
        bOK = ReadFile( hCom, &pBuffer[nTotalNbBytesRead], 1, &nNbBytesRead, NULL );
        if ( bOK && nNbBytesRead != 0 )
        {
          nTotalNbBytesRead++;    
          if ( pBuffer[nTotalNbBytesRead - 1] == '>' )      // last char
            bDone = TRUE;
        }
        else
          bDone = TRUE;
      }
    }           // end if bOK 
  }             // End RD_FEEDBACK    
  

  if ( nTotalNbBytesRead > 0 )
  {
    if ( bStoreToFifo )
      DCPU_CopyBufferToFifo( (char *) pBuffer, 
                             (int) nTotalNbBytesRead, 
                             cSeparator,
                             FALSE,     // bWrite
                             TRUE,      // bNewCommand
                             nPhase == RD_RESPONSE ? CMD_NORMAL: CMD_FEEDBACK );    
    *pNbBytesRead = nTotalNbBytesRead;
    nTotalNbBytesRead++;
    pBuffer[nTotalNbBytesRead] = '\0';    // for direct string display
  }
  else
    *pNbBytesRead = nTotalNbBytesRead;  // also copy if 0

  return bOK;
}

/*======================================================================*/
/*      DCPU_CopyBufferToFifo                                           */
/*      copies nNbChars bytes from address pOrgPtr, to the debug FIFO,  */
/*      and in the receive FIFO.                                        */
/*      Additionally, if flag bInsertDumpSeparator is set, then a       */
/*      one character separator (depending on current phase) is         */
/*      inserted into the fifo, allowing to visualize the packetting of */
/*      READs.                                                          */

void            
DCPU_CopyBufferToFifo( char *pOrgPtr, int nNbBytes, char cSeparator,
                       BOOL bWrite, BOOL bNewCmd, CMD_TYPE nCmdType )
{
  int                   nByte, nStartByte;
  char                  *Ptr;
  char                  sMessage[2 * MAX_MSG_LEN];
  char                  sString[10];

  if ( pDebugFile != NULL )
  {
    if ( *pOrgPtr == '<' )      // string
      sprintf( sMessage , "%d bytes |%s| ", nNbBytes, pOrgPtr );
    else
    {
      sprintf( sMessage, "%d bytes |", nNbBytes );
      for ( nByte = 0; nByte < nNbBytes ; nByte++ )
      {
        sprintf( sString, "%02X ", pOrgPtr[nByte] );
        strcat( sMessage, sString );
      }
      strcat( sMessage, "|" );
    }
    DCPU_PrintDebugMessage( sMessage );
  }

  if ( bInsertDumpSeparators && cSeparator != 0x00 )
  {
    nRcvFifoWrPtr = nRcvFifoWrPtr + 1;
    if ( nRcvFifoWrPtr >= RCV_FIFO_SIZE )
      nRcvFifoWrPtr = 0;
    cRcvFifo[nRcvFifoWrPtr] = cSeparator;
  }

  Ptr = pOrgPtr;

  if ( bWrite || bNewCmd )
  {
    nCmdFifoWrPtr = nCmdFifoWrPtr + 1;
    if ( nCmdFifoWrPtr >= MAX_DBG_BUFFER_NB_LINES )
      nCmdFifoWrPtr = 0;
    xCmdFifo[nCmdFifoWrPtr].lTimeTick = xDccCtxt.xServerDesc.lCurTimeTick;
    xCmdFifo[nCmdFifoWrPtr].bWrite = bWrite;
    xCmdFifo[nCmdFifoWrPtr].bBroadcast = FALSE;
    xCmdFifo[nCmdFifoWrPtr].bResponsePatched = FALSE;
    xCmdFifo[nCmdFifoWrPtr].nNbBytes = 0;
    xCmdFifo[nCmdFifoWrPtr].nCommandType = nCmdType;
  }

  nStartByte = xCmdFifo[nCmdFifoWrPtr].nNbBytes;

  for ( nByte = 0; nByte < nNbBytes; nByte++ )
  {
    nRcvFifoWrPtr = nRcvFifoWrPtr + 1;
    if ( nRcvFifoWrPtr >= RCV_FIFO_SIZE )
      nRcvFifoWrPtr = 0;
    cRcvFifo[nRcvFifoWrPtr] = *Ptr;

    xCmdFifo[nCmdFifoWrPtr].cBuffer[nStartByte + nByte] = *Ptr;

    Ptr++;
  }

  xCmdFifo[nCmdFifoWrPtr].nNbBytes += nNbBytes;

  if ( bNewCmd )
    nTotalCmdNb++;

  if ( nTotalCmdNb - nLastFlushCmdIndex >= MAX_DBG_BUFFER_NB_LINES - 8 )
  {
    nLastFlushCmdIndex = nTotalCmdNb;
    DCPU_DumpCmdFifoToFile( pDumpFile );
  }
}
                                         
/*======================================================================*/
/*      DCPU_RemoveLastCmdFifoEntry                                     */
/*      This function is invoked any time the general event poll (C8),  */
/*      yields response 00 (no event).                                  */
/*      It removes the two last entries (command C8 + response 00 in    */
/*      command Fifo.                                                   */

void
DCPU_RemoveLastCmdFifoEntry( void )
{
  nCmdFifoWrPtr = nCmdFifoWrPtr - 1;
  if ( nCmdFifoWrPtr < 0 )
   nCmdFifoWrPtr = MAX_DBG_BUFFER_NB_LINES - 1;

  nCmdFifoWrPtr = nCmdFifoWrPtr - 1;
  if ( nCmdFifoWrPtr < 0 )
   nCmdFifoWrPtr = MAX_DBG_BUFFER_NB_LINES - 1;

  nTotalCmdNb -= 2;
}

/*======================================================================*/
/*      DCPU_UpdateLastCmdFifoEntry                                     */
/*      This function is invoked from DCPT_ScanResponse.                */
/*      Used in the case of LOCAL commands only.                        */
/*      It patches the last SenD:RECV pair in the case of local cmds.   */
/*      It searches back until it finds the last send command,          */
/*                                                                      */
/*      NOTE: since CMD_REFRESH commands are not stored, NO UPDATE      */
/*      should be done for these commands.                              */

void
DCPU_UpdateLastCmdFifoEntry( DCC_COMMAND_CTXT *pSndCommandCtxt )
{
  int           nPtr;
  DCC_COMMAND   *pRcvCommand;
  BOOL          bEnd = FALSE;

  if ( pSndCommandCtxt->nCommandType != CMD_LOCAL )
    return;

  nPtr = nCmdFifoWrPtr;
  while ( !xCmdFifo[nPtr].bWrite )
  {
    nPtr--;
    if ( nPtr < 0 )
      nPtr = MAX_DBG_BUFFER_NB_LINES - 1;
  }

  while ( !bEnd )
  {
    xCmdFifo[nPtr].nCommandType = CMD_LOCAL;
    if ( nPtr != nCmdFifoWrPtr )
    {  
      nPtr++;
      if ( nPtr >= MAX_DBG_BUFFER_NB_LINES )
        nPtr = 0;
    }
    else
      bEnd = TRUE;
  }

  /* in the case of no event: remove entry in command fifo */
  pRcvCommand = &pSndCommandCtxt->xRcvCommand;

}

/*======================================================================*/
/*      DCPU_DumpDbgFifo                                                */
/*      Display contents of 1024  characters in FIFO.                     */

void            
DCPU_DumpDbgFifo( HWND hWnd )
{
  char                  sMessage[4 * MAX_LONG_MSG_LEN] = "\0" ;
  char                  sLine[RCV_FIFO_SIZE / 32][2 * MAX_STR_LEN];
  char                  sElem[2 * MAX_STR_LEN];
  int                   nLine;
  int                   nByte;
  int                   nRdPtr;

  nRdPtr = (nRcvFifoWrPtr - RCV_FIFO_SIZE + 1);
  if ( nRdPtr < 0 )
    nRdPtr += RCV_FIFO_SIZE;
  for ( nLine = 0; nLine < RCV_FIFO_SIZE / 32 ; nLine++ )
  {
    sLine[nLine][0] = '\0';
    for ( nByte = 0; nByte < 32; nByte++ )
    {
      sprintf( sElem, "%02X ", (unsigned char) cRcvFifo[nRdPtr] );
      strcat( sLine[nLine], sElem );
      nRdPtr++;
      if ( nRdPtr >= RCV_FIFO_SIZE )
        nRdPtr = 0;
    }
    strcat( sLine[nLine], "\n" );
    strcat( sMessage, sLine[nLine] );
  }

  MessageBox( hWnd, sMessage, "RCV FIFO DUMP", 
    MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );

}

/*======================================================================*/
/*      DCPU_DumpDbgFifoToFile                                          */
/*      Prints contents of 256 characters in FIFO, to pFile             */

void            
DCPU_DumpDbgFifoToFile( FILE *pFile )
{
  char                  sLine[MAX_STR_LEN];
  char                  sElem[MAX_STR_LEN];
  int                   nLine;
  int                   nByte;
  int                   nRdPtr;

  if ( pFile == NULL ) return;

  fprintf( pFile, "\nRCV FIFO DUMP\n" );

  nRdPtr = (nRcvFifoWrPtr - RCV_DUMP_SIZE + 1);
  if ( nRdPtr < 0 )
    nRdPtr += RCV_FIFO_SIZE;
  for ( nLine = 0; nLine < RCV_DUMP_SIZE / 16 ; nLine++ )
  {
    sLine[0] = '\0';
    for ( nByte = 0; nByte < 16 ; nByte++ )
    {
      sprintf( sElem, "%02X ", (unsigned char) cRcvFifo[nRdPtr] );
      strcat( sLine, sElem );
      nRdPtr++;
      if ( nRdPtr >= RCV_FIFO_SIZE )
        nRdPtr = 0;
    }
    fprintf( pFile, "%s\n" , sLine );
  }
  fflush( pFile );
}

/*======================================================================*/
/*      DCPU_DumpCmdFifo                                                */
/*      Writes 32 first entries of xCmdFifo. to hWnd, or pFile if not   */
/*      NULL.                                                           */
/*      NOTE: CDM_REFRESH command are not stored and thus not handled   */
/*      by this function.                                               */

void            
DCPU_DumpCmdFifo( HWND hWnd, FILE *pFile, BOOL bError )
{
  char                  sMessage[8 * MAX_LONG_MSG_LEN] = "\0" ;
  char                  sLine[MAX_DBG_BUFFER_NB_LINES][256];
  char                  sElem[256];
  int                   nLine;
  int                   nMaxNbLines;
  int                   nByte;
  int                   nRdPtr;
  char                  sString[256];
  int                   nStringIndex;

  nMaxNbLines = (pFile == NULL)? 
                        MAX_DBG_DISPLAY_NB_LINES : MAX_DBG_BUFFER_NB_LINES;

  nRdPtr = (nCmdFifoWrPtr - nMaxNbLines + 1);
  if ( nRdPtr < 0 )
    nRdPtr += MAX_DBG_BUFFER_NB_LINES;

  for ( nLine = 0; nLine < nMaxNbLines ; nLine++ )
  {
    if ( xCmdFifo[nRdPtr].bWrite )
      sprintf( sLine[nLine], "%05ld %s: ", 
        xCmdFifo[nRdPtr].lTimeTick, 
        sCommandPrefix[xCmdFifo[nRdPtr].nCommandType][0] );

    else
      sprintf( sLine[nLine], "%05ld %s: ", 
        xCmdFifo[nRdPtr].lTimeTick, 
        sCommandPrefix[xCmdFifo[nRdPtr].nCommandType][1] );

    strcpy( sString, "   (" );
    nStringIndex = strlen( sString );

    for ( nByte = 0; 
          nByte < xCmdFifo[nRdPtr].nNbBytes; 
          nByte++ )
    {

      // do not print hexa bytes if nB bytes > 8 
      if ( xCmdFifo[nRdPtr].nNbBytes <= 8 )
      {
        sprintf( sElem, "%02X ", xCmdFifo[nRdPtr].cBuffer[nByte] );
        strcat( sLine[nLine], sElem );
      }

      if ( xCmdFifo[nRdPtr].nCommandType == CMD_NORMAL ||
           xCmdFifo[nRdPtr].nCommandType == CMD_LOCAL ||
           xCmdFifo[nRdPtr].nCommandType == CMD_FEEDBACK )
      {
        if ( xCmdFifo[nRdPtr].cBuffer[nByte] != '\r' &&
             xCmdFifo[nRdPtr].cBuffer[nByte] != '\n' )
          sString[nStringIndex++] = xCmdFifo[nRdPtr].cBuffer[nByte];
        else
          sString[nStringIndex++] = ' ';
      }
    }
    sString[nStringIndex] = '\0';       /* for strcat */

    switch( xCmdFifo[nRdPtr].nCommandType )
    {
    case CMD_LOCAL_FEEDBACK:
      strcpy( sElem, DCPU_RcvCmdMnemonicStr( 
                        (RCV_CMD_MNEMONIC) xCmdFifo[nRdPtr].cBuffer[0] ) );
      strcat( sString, sElem );
      nStringIndex += strlen( sElem );
      break;

    case CMD_STANDALONE:
      if ( xCmdFifo[nRdPtr].bWrite )
      {
        strcpy( sElem, DCPU_SndCmdMnemonicStr( 
                        (SND_CMD_MNEMONIC) xCmdFifo[nRdPtr].cBuffer[0] ) );
        strcat( sString, sElem );
        nStringIndex += strlen( sElem );
      }
      else      /* the 3 bytes contain: RCV command, completion code, error */
      {
        sprintf( sElem, "%s/%s/%s", 
                 DCPU_RcvCmdMnemonicStr( 
                        (RCV_CMD_MNEMONIC) xCmdFifo[nRdPtr].cBuffer[0] ),
                 DCPU_CompletionCodeStr( 
                        (DCC_COMPLETION_CODE) xCmdFifo[nRdPtr].cBuffer[1] ),
                 DCPU_RdErrorStr( 
                        (RD_ERROR) xCmdFifo[nRdPtr].cBuffer[2] ) );
 
        strcat( sString, sElem );
        nStringIndex += strlen( sElem );
      }
      break;
      

    default:    /* do nothing else */
      break;
    }

    sString[nStringIndex++] = ')';
    sString[nStringIndex++] = '\0';
    strcat( sLine[nLine], sString );
    strcat( sLine[nLine], "\n" );

    if ( pFile == NULL )
      strcat( sMessage, sLine[nLine] );

    nRdPtr++;
    if ( nRdPtr >= MAX_DBG_BUFFER_NB_LINES )
      nRdPtr = 0;
  }

  if ( pFile == NULL )
  {
    MessageBox( hWnd, sMessage, "CMD FIFO DUMP", 
      MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
  }
  else
  {
    if ( !bError )
      fprintf( pFile , "\n============================================\n") ;
    else
      fprintf( pFile , "\n=====================ERROR==================\n") ;

    for ( nLine = 0; nLine < nMaxNbLines ; nLine++ )
      fprintf( pFile, "%s", sLine[nLine] );

    fprintf( pFile , "============================================\n" );
    fflush( pFile );    
  }
}

/*======================================================================*/
/*      DCPU_DumpCmdFifoToFile                                          */
/*      Writes all entries of xCmdFifo. to pFile                        */

void            
DCPU_DumpCmdFifoToFile( FILE *pFile )
{

  if ( pFile == NULL ) return;

  fprintf( pFile , "\n=================CMD DUMP===================\n") ;

  DCPU_DumpCmdFifo( NULL, pFile, FALSE );

  fprintf( pFile , "============================================\n" );
  fflush( pFile );    

}

/*======================================================================*/
/*      DCPU_PrintMessageToFile                                         */

static nNbLines = 0;

void    
DCPU_PrintMessageToFile( FILE  *pFile, char sMessage[] )
{
  BOOL          bEnablePrint = TRUE;

  if ( bEnablePrint && pFile != NULL && nNbLines < 10000 )
  {
    fprintf( pFile, "time: %ld - %s\n", 
      xDccCtxt.xServerDesc.lCurTimeTick, sMessage );
    fflush( pFile );
    nNbLines++;
  }
}

/*======================================================================*/
/*      DCPU_ClearCmdTag                                                */

void
DCPU_ClearCmdTag( pDCC_CMD_TAG pCmdTag )
{
  pCmdTag->pObjPtr = NULL;
  pCmdTag->nCmdIdx = 0;
  pCmdTag->lTimeTickSchedule = 0L;
  pCmdTag->lTimeTickDone = 0L;
  pCmdTag->nAddress = 0;
}
                                         
/*======================================================================*/
/*      DCPU_CopySndCommand                                             */
/*      This function copies the send command of index nCommand to      */
/*      *pCommand ( destination in xDccCtxt.xSndMsgFifo ).              */
/*      The body of the command is copied from xSndCommand[nCommand].   */
/*      The data values are copied from *pDataArray.                    */

void
DCPU_CopySndCommand(    SND_CMD_MNEMONIC        nCommand, 
                        int                     nNbArgs,
                        int                     *pDataArray,
                        DCC_COMMAND             *pCommand )
{
  int                   nByte;
  int                   i;
  char                  sMessage[64];

  pCommand->cHeaderCode         = xSendCommand[nCommand].cHeaderCode;
  pCommand->cCommandCode        = xSendCommand[nCommand].cCommandCode;
  pCommand->nNbArgs             = xSendCommand[nCommand].nNbArgs;
  pCommand->nLatency            = xSendCommand[nCommand].nLatency;
  pCommand->nResponse[0]        = xSendCommand[nCommand].nResponse[0];
  strcpy( pCommand->sExpectedRespHeader, xSendCommand[nCommand].sExpectedRespHeader );

  /* pCommand->nNbBytes not set at this level. Will be set by DCPT_SendCommand*/
  for ( nByte = 0; nByte < nNbArgs; nByte++ )
  {
    pCommand->xDataByte[nByte] = pDataArray[nByte];
  }
}

/*======================================================================*/
/*      DCPU_XXXAccessoryAddress                                        */
/*      These 2 functions the 0-based address of accessory whose        */
/*      address is in first byte xDataByte[0].cValue,                   */
/*      and nibble in xDataByte[1].cValue.                              */
/*      The address part is common to both Rcv and Snd.                 */
/*      On the contrary, the nibble bit is bit 4 of xDataByte[1].cValue */
/*      in the case of the response (Rcv), and is bit 0 in the case of  */
/*      the send command.                                               */

int
DCPU_RcvAccessoryAddress( CMD_DATA xDataByte[] )
{
  int           nAddress, nNibble;

  nAddress = 8 * ( ( int) (xDataByte[0].cValue) );
  nNibble = (int) ( xDataByte[1].cValue & (unsigned char) 0x10 );
  if ( nNibble != 0 )
    nNibble = 4;
  nAddress += nNibble;

  return nAddress;
}

int
DCPU_SndAccessoryAddress( CMD_DATA xDataByte[] )
{
  int           nAddress, nNibble;

  nAddress = 8 * ( ( int) (xDataByte[0].cValue) );
  nNibble = (int) ( xDataByte[1].cValue & (unsigned char) 1 );
  if ( nNibble != 0 )
    nNibble = 4;
  nAddress += nNibble;

  return nAddress;
}


/*======================================================================*/
/*                      DCC_COMMAND_CTXT structure hangling             */
/*======================================================================*/

/*======================================================================*/
/*      DCPU_CreateCommandCtxt                                          */

pDCC_COMMAND_CTXT
DCPU_CreateCommandCtxt( void )
{
  pDCC_COMMAND_CTXT     pCmdCtxt;

  pCmdCtxt = 
    ( pDCC_COMMAND_CTXT ) calloc( 1, sizeof( struct S_DCC_COMMAND_CTXT ) );

  return pCmdCtxt;
}

/*======================================================================*/
/*      DCPU_ClearCommandCtxt                                           */
/*      Resets all fields of pCmdCtxt                                   */

void
DCPU_ClearCommandCtxt( pDCC_COMMAND_CTXT pCmdCtxt )
{
  DCPU_ClearCommand( &pCmdCtxt->xSndCommand );
  DCPU_ClearCommand( &pCmdCtxt->xRcvCommand );
  pCmdCtxt->nSndSeqState =      DCC_SEQ_NONE;
  pCmdCtxt->bDisplay =          FALSE;
  pCmdCtxt->nCompletionCode =   DGI_COMP_NONE;
  pCmdCtxt->nRdErrorCode =      RD_ERR_NONE;
  pCmdCtxt->nRdCommand =        RCV_NONE;
  pCmdCtxt->pPrev =             NULL;
  pCmdCtxt->pNext =             NULL;
  pCmdCtxt->xCmdTag.pObjPtr =   NULL;
  pCmdCtxt->xCmdTag.nCmdIdx =   0;
  pCmdCtxt->xCmdTag.lTimeTickSchedule = 0L;
  pCmdCtxt->xCmdTag.lTimeTickDone =     0L;
}


   
/*======================================================================*/
/*      DCPU_CopyCommandCtxt                                            */
/*      This function copies the contents of pSrcCmdCtxt to             */
/*      pDstCmdCtxt, EXCEPT for pPrev and pNext.                        */

void
DCPU_CopyCommandCtxt(   pDCC_COMMAND_CTXT       pDstCmdCtxt, 
                        pDCC_COMMAND_CTXT       pSrcCmdCtxt )
{
  pDCC_COMMAND_CTXT             pPrev, pNext;

  pPrev = pDstCmdCtxt->pPrev;
  pNext = pDstCmdCtxt->pNext;
  *pDstCmdCtxt = *pSrcCmdCtxt;
  pDstCmdCtxt->pPrev = pPrev;
  pDstCmdCtxt->pNext = pNext;
}

/*======================================================================*/
/*      DCPU_MapResponseToGeneric                                       */
/*      Post processor invoked in order to format data generated by     */
/*      DCPU_ParseCommandCtxt, into DGI compatible structure.           */
/*      Invoked in the case of RESPONSES.                               */

void    
DCPU_MapResponseToGeneric( 
                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                pDCC_COMMAND_CTXT       pCmdCtxt,
                RCV_CMD_MNEMONIC        *pCommand )
{
  char                  cBuffer[10];

  if ( pDebugFile != NULL )
  {
    DCPU_PrintDebugMessage( "MapResponseToGeneric" );
  }

  pGenericRcvData->nNbDataItems = 1;

  switch( *pCommand )
  {
  case RCV_ACCESSORY_INFO: 
    DCPU_MapAccessoryInfoToGeneric( pGenericRcvData );     
    break;

  case RCV_DETECTOR_INFO:
    DCPU_MapDetectorInfoToGeneric( pGenericRcvData );     
    break;

  case RCV_NORMAL_LOCO_INFO:
    DCPU_MapLocoInfoToGeneric( pGenericRcvData );     
    break;

  case RCV_SRVC_MODE_RESP_DIRECT_CV:
    DCPU_MapCVInfoToGeneric( pGenericRcvData, pCmdCtxt, pCommand ); 
    break;

  case RCV_CMDSTATION_STATUS:
    DCPU_MapCSStatusToGeneric( pGenericRcvData, pCmdCtxt );    
    break;

  case RCV_NORMAL_OPER_RESUMED:
  case RCV_EMERGENCY_STOP:
  case RCV_TRACK_POWER_OFF:
  case RCV_SERVICE_MODE_ENTRY:          /* NO further data to retrieve */
    break;

  }

  /* following intended for viewinf standalone responses */
  cBuffer[0] = *pCommand;
  cBuffer[1] = pCmdCtxt->nCompletionCode;
  cBuffer[2] = pCmdCtxt->nRdErrorCode;
  DCPU_CopyBufferToFifo( cBuffer, 3, (char) 0xDD,
                         FALSE,         /* bWrite */
                         TRUE,          /* bNewCmd */
                         pCmdCtxt->nCommandType );
}

/*======================================================================*/
/*      DCPU_MapAccessoryInfoToGeneric                                  */
/*      Subfunction of DCPU_MapResponseToGeneric, in the case of        */
/*      accessory (turnout) info.                                       */
/*                                                                      */
/*      - pGenericRcvData->xDataItem[0].nAddress  : TO address          */
/*      - pGenericRcvData->xDataItem[0].hData[0] :                      */
/*        2 bits: bOutput2 (bit 1) bOutput (bit 0)                      */
/*                                                                      */
/*      IMPORTANT NOTE: the address in CMD_TAG s the real acceessory    */
/*      address, starting from 1.                                       */
/*      On the other hand, the address to be written in pGenericRcvData */
/*      is the BASE 0 address of the first accessory in nibble.         */

void    
DCPU_MapAccessoryInfoToGeneric( pDGI_GENERIC_RCV_DATA   pGenericRcvData )
{
  int                   nAddress;

  /* return address base 0 (address in Cmdtag starts from 1) */
  pGenericRcvData->xDataItem[0].nAddress = 
    pGenericRcvData->xCmdTag.nAddress - 1;

  DCPS_GetAccessoryInfo( pGenericRcvData, nAddress );
}

/*======================================================================*/
/*      DCPU_MapDetectorInfoToGeneric                                   */
/*      Subfunction of DCPU_MapResponseToGeneric, in the case of        */
/*      detector info.                                                  */
/*                                                                      */
/*      - pGenericRcvData->xDataItem[0].nAddress  :                     */
/*                              address of 1st detector in nibble.      */
/*      - pGenericRcvData->xDataItem[0].hData[0] : nibble value.        */

void    
DCPU_MapDetectorInfoToGeneric( pDGI_GENERIC_RCV_DATA pGenericRcvData )
{
  int                   nNibbleAddress;

  // address of 1st detector in nibble
  pGenericRcvData->xDataItem[0].nAddress = 
                        4 * (( pGenericRcvData->xCmdTag.nAddress - 1 ) / 4 );
  // nibble value
  nNibbleAddress = (pGenericRcvData->xCmdTag.nAddress - 1) / 4;

  pGenericRcvData->xDataItem[0].hData[0] = 
        (unsigned short) (xS88State[nNibbleAddress] & ( (unsigned char) 0x0F ));
}

/*======================================================================*/
/*      DCPU_MapLocoInfoToGeneric                                       */
/*      Subfunction of DCPU_MapResponseToGeneric, in the complex case of*/
/*      loco info.                                                      */
/*                                                                      */
/*      After parsing:                                                  */
/*      xData[0]:       nSpeed                                          */
/*      xData[1]:       nF0F4                                           */
/*                                                                      */
/*      - pGenericRcvData->xDataItem[0].hData[0] : speed                */
/*      - pGenericRcvData->xDataItem[0].hData[1] : F0F12                */
/*      - pGenericRcvData->xDataItem[0].hData[2] : function mask        */
/*                              (bits in hdata[1] relevant or not)      */
/*      - pGenericRcvData->xDataItem[0].hData[3] : status.              */
/*        bit 0: validity of step mode.                                 */
/*        bit 1: validity of remote control bit                         */
/*        bits 8-9-10: step mode.                                       */
/*        bit 11: remote control bit                                    */
/*        Everything at 0 because nothing valid.                        */


void    
DCPU_MapLocoInfoToGeneric(      pDGI_GENERIC_RCV_DATA   pGenericRcvData )
{ 
  int                   nAddress;

  pGenericRcvData->xDataItem[0].nAddress =  pGenericRcvData->xCmdTag.nAddress;

  DCPS_GetLocoInfo( pGenericRcvData, nAddress );
}

/*======================================================================*/
/*      DCPU_MapCVInfoToGeneric                                         */
/*      Subfunction of DCPU_MapResponseToGeneric, in the complex case of*/
/*      CV info.                                                        */
/*                                                                      */
/*      pCmdCtxt->xRcvCommand.xDataByte[0]       CV Value               */
/*                                                                      */
/*      - pGenericRcvData->xDataItem[0].nAddress : copied from TAG      */
/*      - pGenericRcvData->xDataItem[0].hData[0] : CV value             */


void    
DCPU_MapCVInfoToGeneric( pDGI_GENERIC_RCV_DATA  pGenericRcvData,
                        DCC_COMMAND_CTXT        *pCmdCtxt,
                        RCV_CMD_MNEMONIC        *pCommand )
{
  int                   nAddress;
  unsigned short        hValue;
  char                  sMessage[64];

/*
  sprintf( sMessage, "%02X %02X %02X" ,
        (unsigned char) xData[0],(unsigned char) xData[1] ,(unsigned char) xData[2] );
    MessageBox( hMainWnd, sMessage, "", 
        MB_APPLMODAL | MB_OK | MB_ICONINFORMATION );
*/

  pGenericRcvData->xDataItem[0].nAddress = pGenericRcvData->xCmdTag.nAddress;

  switch( pCmdCtxt->nRdErrorCode )
  {
  case RD_ERR_NONE:
    pGenericRcvData->xDataItem[0].hData[0] =
        (unsigned int) pCmdCtxt->xRcvCommand.xDataByte[0];
    *pCommand = RCV_SRVC_MODE_RESP_DIRECT_CV;
    break;

  default:
    *pCommand = RCV_PROG_DATA_BYTE_NOT_FOUND;
    break;
  }
}


/*======================================================================*/
/*      DCPU_MapCSStatusToGeneric                                       */
/*      Subfunction of DCPU_MapResponseToGeneric, in the complex case of*/
/*      Command station status.                                         */
/*                                                                      */
/*      Status stored in hData[0]                                       */
/*      - bit 0 (1) means emergency stop.                               */
/*      - bit 1 (1) means power on, 0 power off.                        */
/*                                                                      */
/*      - pGenericRcvData->xDataItem[0].hData[0] : status mapped to     */
/*        standard generic (Lenz)                                       */


void    
DCPU_MapCSStatusToGeneric(      pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                DCC_COMMAND_CTXT        *pCmdCtxt  )
{ 
  int                   nAddress;
  unsigned short        hValue = 0;

  if (xDccCtxt.xServerDesc.bEmergencyStop )
    hValue += ( 1 << DGI_CS_STATUS_STOP_BIT );

  if ( xDccCtxt.xServerDesc.bPowerOff )
    hValue += ( 1 << DGI_CS_STATUS_POWER_OFF_BIT );

  pGenericRcvData->xDataItem[0].hData[0] = hValue;
}

/*======================================================================*/
/*      DCCU_MapMessageToGeneric                                        */
/*      Invoked in the case of UNEXPECTED RESPONSES.                    */
/*      Maps the info stored in local arrays, to pGenericRcvData (std   */
/*      communication strcuture for CDM-                                */

void    
DCPU_MapMessageToGeneric( 
                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                pDCC_COMMAND_CTXT       pCmdCtxt,
                RCV_CMD_MNEMONIC        *pCommand )
{
  switch( *pCommand )
  {
  case RCV_FEEDBACK_BROADCAST:      
    DCPU_MapFeedbackBroadcastToGeneric( pGenericRcvData );
    break;

  }

}

/*======================================================================*/
/*      DCPU_MapFeedbackBroadcastToGeneric                              */
/*      Subfunction of DCPU_MapMessageToGeneric, in the complex case of */
/*      a feedback broadcast.                                           */
/*                                                                      */
/*      In the case of a RCV_DETECTOR_EVENT_REPORT,                     */                   
/*                                                                      */
/*      - pGenericRcvData->nNbDataItems is set to number of changes     */
/*      - pGenericRcvData->xDataItem[i].nAddress is set to address of   */
/*        first detector in nibble - 1 (base 0).                        */
/*      - pGenericRcvData->xDataItem[i].hData[0]: 4 bit nibble value.   */
/*                                                                      */

void    
DCPU_MapFeedbackBroadcastToGeneric( pDGI_GENERIC_RCV_DATA pGenericRcvData )
{
  int                   nNibbleIndex;
  int                   nNbChanges = 0;
  int                   nGRDIndex;
  int                   nMask;
  BOOL                  bOK = TRUE;

  nGRDIndex = 0;

  for ( nNibbleIndex = 0; 
        nNibbleIndex < (MAX_NB_DETECTORS / 4) && bOK ; 
        nNibbleIndex++ )
  {
    nMask = 1 << nNibbleIndex;
    if ( ( nS88NibbleChanges & nMask ) != 0 )
    {
      if ( nNbChanges < DGI_MAX_DATA_ITEMS )
      {
        nNbChanges++;
        pGenericRcvData->xDataItem[nGRDIndex].nAddress = 4 * nNibbleIndex;
        pGenericRcvData->xDataItem[nGRDIndex].hData[0] = 
                        (xS88State[nNibbleIndex] & (unsigned char) 0x0F );
        
        nGRDIndex++;
        // erase change bit
        nS88NibbleChanges &= ~nMask;
      }
      else      /* more than DGI_MAX_DATA_ITEMS: overload */
        bOK = FALSE;    // leave for next time        
    }
  }

  // and finally set number of changes
  pGenericRcvData->nNbDataItems = nNbChanges;

}

/*======================================================================*/
/*      DCPU_ClearCommand                                               */
/*      Resets all fields of pCommand                                   */

void
DCPU_ClearCommand( pDCC_COMMAND pCommand )
{
  int           i;

  pCommand->cHeaderCode =       (unsigned char) 0;
  pCommand->cCommandCode =      (unsigned char) 0;
  pCommand->nNbArgs     =       0;
  pCommand->nLatency    =       0;
  for ( i = 0; i < MAX_NB_RESPONSES; i++ )
    pCommand->nResponse[i] =    RCV_NONE;

  for ( i = 0; i < MAX_NB_CMD_DATA; i++ )
  {
    pCommand->xDataByte[i] = 0;
  }
}

/*======================================================================*/
/*      DCPU_NewCommandCtxt                                             */
/*      This function appends a new command to list pCmdEnv->pCmdList.  */
/*      If there is a free CmdCtxt in xDccCtxt.pFreeCmdCtxtList, it is  */
/*      detached from head of that list, and cleared.                   */
/*      Otherwise, DCPU_CreateCommandCtxt is invoked.                   */

pDCC_COMMAND_CTXT
DCPU_NewCommandCtxt( pDCC_CMD_CTXT_ENV pCmdEnv, CMD_PRIORITY nPriority )
{
  pDCC_COMMAND_CTXT     pCmdCtxt, pCmdCtxt2;
  pDCC_COMMAND_CTXT     pFirstCmdCtxt, pLastCmdCtxt, pPrevCmdCtxt;
  BOOL                  bFound;
  BOOL                  bNew;
  char                  sMessage[2 * MAX_STR_LEN];

  if ( xDccCtxt.nNbFreeCmdCtxt > 0 )
  {
    pCmdCtxt = xDccCtxt.pFreeCmdCtxtList;
    xDccCtxt.pFreeCmdCtxtList = xDccCtxt.pFreeCmdCtxtList->pNext;
    DCPU_ClearCommandCtxt( pCmdCtxt );
    xDccCtxt.nNbFreeCmdCtxt--;
    bNew = FALSE;
  }
  else
  {
    pCmdCtxt = DCPU_CreateCommandCtxt();
    bNew = TRUE;
  }

  pCmdCtxt->nPriority = nPriority;
  pCmdCtxt->lIndex = ++lCurrentCommandCtxtIndex;

  if ( bLowLevelDebug && (pCmdEnv == &xDccCtxt.xSndCmd ) )
  {
    sprintf( sMessage, 
          "DCPU_NewCommandCtxt: send cmd_ctxt %ld cmd type %d cmd pri %d New = %d nb free ctxt %d send ctxt nb %d" , 
          pCmdCtxt->lIndex, pCmdCtxt->nCommandType, nPriority, 
          bNew, xDccCtxt.nNbFreeCmdCtxt, xDccCtxt.xSndCmd.nNbCmd  );
    DCPU_PrintMessageToFile( pDumpFile, sMessage );
  }

  /*
  if ( nPriority < PRI_3 )
  {
    * insert at end of end of list pCmdEnv->pCmdList *
    if ( pCmdEnv->nNbCmd == 0 )
      pCmdEnv->pCmdCtxtList = pCmdCtxt;
    else
    {
      pCmdEnv->pCmdWrPtr->pNext = pCmdCtxt;
      pCmdCtxt->pPrev = pCmdEnv->pCmdWrPtr;
    }
    pCmdEnv->pCmdWrPtr = pCmdCtxt;
  }
  else  * bUrgent *
  {
    if ( pCmdEnv->nNbCmd == 0 )
    {
      pCmdEnv->pCmdCtxtList = pCmdCtxt;
      pCmdEnv->pCmdWrPtr = pCmdCtxt;
    }

    else if ( pCmdEnv->pCmdRdPtr == NULL )   * insert in first pos. *
    {
      pCmdCtxt->pNext = pCmdEnv->pCmdCtxtList;
      pCmdEnv->pCmdCtxtList = pCmdCtxt;
    }

    else        * insert after last command processed *
    {
      pLastCmdCtxt = pCmdEnv->pCmdRdPtr;
      pCmdCtxt->pNext = pLastCmdCtxt->pNext;
      pLastCmdCtxt->pNext = pCmdCtxt;
      pCmdCtxt->pPrev = pLastCmdCtxt;
      if ( pCmdCtxt->pNext != NULL )
        pCmdCtxt->pNext->pPrev = pCmdCtxt;
      else      * last command of list *
        pCmdEnv->pCmdWrPtr = pCmdCtxt;
    }  
  }

  */


  if ( pCmdEnv->nNbCmd == 0 )
  {
    pCmdEnv->pCmdCtxtList = pCmdCtxt;
    pCmdEnv->pCmdWrPtr = pCmdCtxt;
  }
  else
  {
    if ( pCmdEnv->pCmdRdPtr == NULL )
      pFirstCmdCtxt = pCmdEnv->pCmdCtxtList;
    else
      pFirstCmdCtxt = pCmdEnv->pCmdRdPtr->pNext;

    pLastCmdCtxt = pCmdEnv->pCmdWrPtr;

    pCmdCtxt2 = pFirstCmdCtxt;
    bFound = FALSE;
    while ( pCmdCtxt2 != NULL && !bFound )
    {
      if ( pCmdCtxt2->nPriority < nPriority )
        bFound = TRUE;
      else
      {
        if ( pCmdCtxt2 == pLastCmdCtxt )
          pCmdCtxt2 = NULL;
        else
          pCmdCtxt2 = pCmdCtxt2->pNext;
      }
    }

    if ( !bFound )      /* insert at end of list */
    {
      pCmdEnv->pCmdWrPtr->pNext = pCmdCtxt;
      pCmdCtxt->pPrev = pCmdEnv->pCmdWrPtr;
      pCmdEnv->pCmdWrPtr = pCmdCtxt;
    }
    else                /* insert before pCmdCtxt2 */
    {
      pPrevCmdCtxt = pCmdCtxt2->pPrev;
      pCmdCtxt2->pPrev = pCmdCtxt;
      pCmdCtxt->pNext = pCmdCtxt2;
      if ( pPrevCmdCtxt != NULL )
      {
        pPrevCmdCtxt->pNext = pCmdCtxt;
        pCmdCtxt->pPrev = pPrevCmdCtxt;
      }
      else      /* head of list */
        pCmdEnv->pCmdCtxtList = pCmdCtxt;
    }
  }


  pCmdEnv->nNbCmd++;
  pCmdEnv->nNbPendingCmd++;
  /* xDccCtxt.lTotalNbCmdSent++; */

  return pCmdCtxt;
}

/*======================================================================*/
/*      DCPU_DetachCommandCtxt                                          */
/*      This function frees the head of list pCmdEnv->pCmdList and      */
/*      adds detached command at end of list xDccCtxt.pFreeCmdCtxtList. */

void
DCPU_DetachCommandCtxt( pDCC_CMD_CTXT_ENV pCmdEnv )
{
  pDCC_COMMAND_CTXT     pCmdCtxt;
  char                  sMessage[2 * MAX_STR_LEN];

  if ( pCmdEnv->nNbCmd > 0 )
  {
    if ( pCmdEnv->pCmdRdPtr == pCmdEnv->pCmdCtxtList )
        pCmdEnv->pCmdRdPtr = NULL;  
    pCmdCtxt = pCmdEnv->pCmdCtxtList;
    pCmdEnv->pCmdCtxtList = pCmdEnv->pCmdCtxtList->pNext;
    if ( pCmdEnv->pCmdCtxtList != NULL )
      pCmdEnv->pCmdCtxtList->pPrev = NULL;
    else
    {
      pCmdEnv->pCmdWrPtr = NULL;
      pCmdEnv->pCmdRdPtr = NULL;
    }
    pCmdEnv->nNbCmd--;

    /* and add at head of free list */
    pCmdCtxt->pNext = NULL;
    pCmdCtxt->pPrev = NULL;

    if ( xDccCtxt.nNbFreeCmdCtxt > 0 )
      pCmdCtxt->pNext = xDccCtxt.pFreeCmdCtxtList;
    xDccCtxt.pFreeCmdCtxtList = pCmdCtxt;
    xDccCtxt.nNbFreeCmdCtxt++;

    if ( bLowLevelDebug && (pCmdEnv == &xDccCtxt.xSndCmd ) )
    {
      sprintf( sMessage, 
          "DCPU_DetachCommandCtxt: send cmd_ctxt %ld cmd %d cmd type %d cmd pri %d nb free ctxt %d send ctxt nb %d" , 
          pCmdCtxt->lIndex, pCmdCtxt->nSndCommand, pCmdCtxt->nCommandType, 
          pCmdCtxt->nPriority, xDccCtxt.nNbFreeCmdCtxt, xDccCtxt.xSndCmd.nNbCmd );
      DCPU_PrintMessageToFile( pDumpFile, sMessage );
    }

  }
}

/*======================================================================*/
/*      DCPU_FreeCommandCtxtEnv                                         */
/*      Frees all DCC_COMMAND_CTXT structures attached to pList.        */

void
DCPU_FreeCommandCtxtEnv( pDCC_CMD_CTXT_ENV pCmdEnv )
{
  pDCC_COMMAND_CTXT     pCmdCtxt, pCmdCtxt2;

  while( pCmdEnv->pCmdCtxtList != NULL )
    DCPU_DetachCommandCtxt( pCmdEnv );

  pCmdEnv->nNbCmd = 0;
  pCmdEnv->nNbPendingCmd = 0;
  pCmdEnv->pCmdRdPtr = NULL;
  pCmdEnv->pCmdWrPtr = NULL;
  pCmdEnv->pCmdCtxtList = NULL;
}

/*======================================================================*/
/*                      DCC FORMAT CONVERSION UTILITIES                 */
/*======================================================================*/

/*======================================================================*/
/*      DCPU_F0F12Byte2Int                                              */
/*      Convert the data contained in bytes FuncA and FuncB, in the     */
/*      order of:                                                       */
/*      - FuncA:          0   0   0 F0 F4 F3 F2 F1                      */
/*      - FuncB:        F12 F11 F10 F9 F8 F7 F6 F5                      */
/*      into an int containing all functions ordered as                 */
/*      F12 F11 ...... F1 F0.                                           */
/*                                                                      */
/*      RETURN: int value built that way.                               */

int
DCPU_F0F12Byte2Int( unsigned char cFuncA, unsigned char cFuncB )
{
  unsigned int          nValue, nF4F1;
  unsigned int          nF0;

  nValue = ((unsigned int) cFuncB) << 8;
  if ( ( cFuncA & ( (unsigned char) 0x10 ) ) != 0 )
    nF0 = 1;
  else
    nF0 = 0;
  nF4F1 = ( ((unsigned int) cFuncA) & 0x0F ) << 1;
  nValue += nF4F1 + nF0;
  
  return nValue;
}

/*======================================================================*/
/*      DCPU_SpeedInt2Byte                                              */
/*      This function formats speed "nSpeed" into the proper byte       */
/*      format according to number of steps (see XpressNet spec.)       */
/*                                                                      */
/*      RETURN: computed byte.                                          */

unsigned char
DCPU_SpeedInt2Byte( int                 nSpeed, 
                    SPEED_STEP_MODE     nSpeedStepMode )
{
  unsigned char         cValue, cBit4;

  if ( nSpeed < 0 )
  {
    nSpeed = -nSpeed;
  }

  switch( nSpeedStepMode )
  {
    case SPEED_STEP_27:
      if ( nSpeed > 27 )
        nSpeed = 27;

    case SPEED_STEP_28:
      if ( nSpeed > 28 )
        nSpeed = 28;
      break;

    case SPEED_STEP_14:
      if ( nSpeed > 14 )
        nSpeed = 14;

    case SPEED_STEP_128:
      if ( nSpeed > 126 )
        nSpeed = 126;
      break;
     
    default: 
      break;

  }

  if ( nSpeed == 0 )
    cValue = 0;
  else if ( nSpeedStepMode == SPEED_STEP_14 ||
            nSpeedStepMode == SPEED_STEP_128 )
    cValue = (unsigned char) (nSpeed + 1);
  else if ( nSpeedStepMode == SPEED_STEP_28 )
  {
    cValue = (unsigned char) (nSpeed + 3);
    /* set LSB to bit 4 */
    cBit4 = (cValue & (unsigned char) 1 ) << 4;
    cValue >>= 1;
    cValue |= cBit4;
  }

  return cValue;
}

/*======================================================================*/
/*                      DATA COM UTILITIES                              */
/*======================================================================*/

/*======================================================================*/
/*      DCPU_GenerateXOR                                                */
/*      This function makes the XOR combination of all bytes            */
/*      of array xByteArray, starting from nStartIndex to               */
/*      nStopIndex.                                                     */
/*                                                                      */
/*      RETURN: XOR byte                                                */

unsigned char
DCPU_GenerateXOR( unsigned char xByteArray[], int nStartIndex, int nStopIndex )
{
  int                   nIndex;
  unsigned char         cXORByte;

  cXORByte = xByteArray[nStartIndex];
  for ( nIndex = nStartIndex + 1; nIndex <= nStopIndex; nIndex++ )
    cXORByte ^= xByteArray[nIndex];

  return cXORByte;
}

/*======================================================================*/
/*      DCPU_SetTimeOut                                                 */
/*      This function sets the time out for data com.                   */
/*      xTimeOut.ReadIntervalTimeout = 20;   usual value of 2 ms is NOT */
/*      enough for DCCpp!!!                                             */                  


void
DCPU_SetTimeOut( HANDLE hCom, DWORD nTimeOut )
{
  COMMTIMEOUTS          xTimeOut;  

  GetCommTimeouts( hCom, &xTimeOut );
  xTimeOut.ReadIntervalTimeout = 20;                     
  xTimeOut.ReadTotalTimeoutMultiplier = 20;      
  xTimeOut.ReadTotalTimeoutConstant = nTimeOut;   
  xTimeOut.WriteTotalTimeoutMultiplier = 20;     // 2ms for any byte
  xTimeOut.WriteTotalTimeoutConstant = 50;  
  SetCommTimeouts( hCom, &xTimeOut ); 
}
  
/*======================================================================*/
/*      DCPU_SkipBlanks                                                 */
/*      This function scans characters of xRcvBuffer, looking for first */
/*      non-blank (and non \n \r ) character.                           */
/*      RETURN: TRUE if non blank character found before EOL,           */
/*              FALSE if not.                                           */

BOOL
DCPU_SkipBlanks( unsigned char xBuffer[], int *pIndex )
{
  int           nIndex;
  char          cChar;
  BOOL          bFound = FALSE;

  nIndex = *pIndex;
  while( ( cChar = xBuffer[nIndex] ) != '\0' && !bFound )
  {
    if ( cChar != ' ' && cChar != '\t' && cChar != '\r' && cChar != '\n' ) 
      bFound = TRUE;
    else
      nIndex++;
  }

  *pIndex = nIndex;

  return bFound;
}

/*======================================================================*/
/*      DCPU_GetNextCharIndex                                           */
/*      This function scans characters of xBuffer, looking for first    */
/*      occurence of any char stored within cExpectedChar.              */
/*      RETURN: TRUE if expected character found before EOL,            */
/*              FALSE if not.                                           */
                                                                        
BOOL
DCPU_GetNextCharIndex( unsigned char    xBuffer[], 
                        unsigned char   xExpectedChar[],
                        int             *pIndex )
{
  int           nIndex;
  char          cChar, cExpectedChar;
  int           nSubIndex;
  BOOL          bFound = FALSE;

  nIndex = *pIndex;
  while( ( cChar = xBuffer[nIndex] ) != '\0' && !bFound )
  {
    nSubIndex = 0;
    while ( ( cExpectedChar = xExpectedChar[nSubIndex] ) != '\0' && !bFound )
    {
      if ( cChar == cExpectedChar ) 
        bFound = TRUE;
      else
        nSubIndex++;
    }
    if ( !bFound )
      nIndex++;
  }

  *pIndex = nIndex;

  return bFound;
}
  
         
/*======================================================================*/
/*      DCPU_IsDecimal                                                  */
/*      Checks if character is a digit.                                 */

BOOL
DCPU_IsDecimal( unsigned char cChar )
{
  if ( cChar >= '0' && cChar <= '9' )
    return TRUE;
  else
    return FALSE;
}

/*======================================================================*/
/*      DCPU_IsDecimal                                                  */
/*      Checks if character is an hexadecimal digit.                    */

BOOL
DCPU_IsHexaDecimal( unsigned char cChar )
{
  if ( ( cChar >= '0' && cChar <= '9' ) ||
       ( cChar >= 'a' && cChar <= 'f' ) ||
       ( cChar >= 'A' && cChar <= 'F' ) )
    return TRUE;
  else
    return FALSE;
}

/*======================================================================*/
/*      DCPU_ExtractDigits                                              */
/*      Extracts nNbDigits from xBuffer, index *pIndex + 1!!            */
/*      only dots are admitted.                                         */
/*      Digits are stroed into xDigitArray.                             */
/*      RETURN: TRUE if correct number of digits extracted,             */
/*              FALSE if not.                                           */

BOOL
DCPU_ExtractDigits(     unsigned char   xRcvBuffer[], 
                        int             *pIndex, 
                        int             nNbDigits, 
                        unsigned char   xDigitArray[] )
{
  int                   nIndex;
  char                  cChar;
  BOOL                  bOK = TRUE;
  BOOL                  bFound = FALSE;
  int                   nExtractIndex;
  

  nIndex = *pIndex + 1;
  DCPU_SkipBlanks( xRcvBuffer, &nIndex );
  nExtractIndex = 0;
  while( ( cChar = xRcvBuffer[nIndex] ) != '\0' && !bFound && bOK )
  {
    if ( cChar >= '0' && cChar <= '9' )
    {
      xDigitArray[nExtractIndex++] = cChar;
      if ( nExtractIndex >= nNbDigits )
        bFound = TRUE;
    }
    else if ( cChar != '.' )
      bOK = FALSE;

    nIndex++;
  }

  *pIndex = nIndex;

  return bFound;
}

/*======================================================================*/
/*      DCPU_SetByte                                                    */

unsigned char
DCPU_SetByte( unsigned char cHigh, unsigned char cLow )
{
  return ( (cHigh << 4) + cLow );
}

/*======================================================================*/
/*      DCPU_Ascii2Hexa                                                 */
/*      Checks and convert the chat string in cBuffer , nNbChars long,  */
/*      to hexadecimal.                                                 */
/*      RETURN TRUE if all cahr hexa, FALSE if not.                     */

BOOL
DCPU_Ascii2Hexa( unsigned char *pBuffer, int nNbChars, unsigned int *pValue )
{
  int           i;
  BOOL          bOK = TRUE;
  char          sMessage[64];

  *pValue = 0;
  /*
  if ( pDebugFile != NULL )
  {
    DCPU_PrintDebugMessage( "DCPU_Ascii2Hexa" );
  }
  */

  for ( i = 0; i < nNbChars && bOK; i++ )
  {
    /*
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage, "start %c i %d value %04X ", pBuffer[i], i , *pValue );
      DCPU_PrintDebugMessage( sMessage );
    }
    */

    if ( !DCPU_IsHexaDecimal( pBuffer[i] ) )
      bOK = FALSE;
    else
    {
      *pValue <<= 4;
      if ( pBuffer[i] >= '0' && pBuffer[i] <= '9' )
        *pValue += ( pBuffer[i] - '0' );
      else if ( pBuffer[i] >= 'a' && pBuffer[i] <= 'f' )
        *pValue += (  pBuffer[i] - 'a' + 10 );
      else if ( pBuffer[i] >= 'A' && pBuffer[i] <= 'F' )
        *pValue += (  pBuffer[i] - 'A' + 10 );
    }
    /*
    if ( pDebugFile != NULL )
    {
      sprintf( sMessage, "end %c i %d value %04X ", pBuffer[i], i , *pValue );
      DCPU_PrintDebugMessage( sMessage );
    }
    */
  }

  return bOK;
}

/*======================================================================*/
/*      DCPU_DisplayVersion                                             */

unsigned char *
DCPU_DisplayVersion(    unsigned char   xByte[], 
                        int             nNbDigits, 
                        int             nAfterDot,
                        unsigned char   xBuffer[] )
{
  int           nBeforeDot;
  int           nNibbleIndex;
  int           nCharIndex;
  int           nByteIndex;
  unsigned char cValue;

  nBeforeDot = nNbDigits - nAfterDot;
  nNibbleIndex = 0;
  nCharIndex = 0;

  while ( nNibbleIndex < nNbDigits )
  {
    if ( nNibbleIndex == nBeforeDot )
      xBuffer[nCharIndex++] = '.';

    nByteIndex = nNibbleIndex / 2;
    if ( nNibbleIndex % 2 == 0 )
      cValue = ( xByte[nByteIndex] >> 4 ) & (unsigned char) 0x0F;
    else
      cValue = xByte[nByteIndex] & (unsigned char) 0x0F;

    xBuffer[nCharIndex] = cValue + '0';

    nNibbleIndex++;
    nCharIndex++;
  }
  
  xBuffer[nCharIndex] = '\0';

  return xBuffer;
}

/*======================================================================*/
/*      DCPU_CommandStationBusy                                         */

BOOL
DCPU_CommandStationBusy( RD_ERROR nRdError )
{
  if ( nRdError == RD_ERR_ALREADY_PROGRAMMING )
    return TRUE;
  else
    return FALSE;
}

/*======================================================================*/
/*                              DEBUG                                   */
/*======================================================================*/

void
DCPU_PrintDebugMessage( char *sDbgMsg )
{
  char          sMessage[128];

  if ( pDebugFile != NULL )
  {
    sprintf( sMessage, "time %ld -> %s\r\n", 
                        xDccCtxt.xServerDesc.lCurTimeTick, sDbgMsg );
    fprintf( pDebugFile, "%s", sMessage );
    fflush( pDebugFile );
  }
}
