/*======================================================================*/
/*                                                                      */
/*      dccpp_util.h                                                    */
/*      Header file of dccpp_util.c                                     */
/*      Module containing miscellaneous utilities related to DCC++      */
/*======================================================================*/
/*      J.P.PILLOU      13/03/2018                                      */
/*======================================================================*/
/*======================================================================*/

#define DBG_FIFO_SIZE   128

char    *DCPU_CompletionCodeStr(
                                DCC_COMPLETION_CODE     nCompletionCode );
                                
char    *DCPU_RcvCmdMnemonicStr(
                                RCV_CMD_MNEMONIC        nRdCommand );

char    *DCPU_RdErrorStr(       RD_ERROR                nRdError );

BOOL    DCPU_WriteFile(         HANDLE                  hCom, 
                                unsigned char           *pBuffer, 
                                DWORD                   nNbBytes, 
                                LPDWORD                 pNbBytesWritten, 
                                BOOL                    bStoreInFifo );


BOOL    DCPU_ReadFile(          HANDLE                  hCom, 
                                unsigned char           *pBuffer, 
                                LPDWORD                 pNbBytesRead,
                                RD_PHASE                nPhase,
                                BOOL                    bStoreToFifo );

void    DCPU_CopyBufferToFifo(  char                    *pOrgPtr,
                                int                     nNbBytes,
                                char                    cSeparator,
                                BOOL                    bWrite,
                                BOOL                    bNewCmd,
                                CMD_TYPE                nCmdType );
                                        

void    DCPU_RemoveLastCmdFifoEntry( void );

void    DCPU_UpdateLastCmdFifoEntry( DCC_COMMAND_CTXT   *pSndCommandCtxt );

void    DCPU_SetLastCmdFifoEntry( CMD_TYPE              nCmdType );

void    DCPU_DumpDbgFifo(       HWND                    hWnd );

void    DCPU_DumpDbgFifoToFile( FILE                    *pFile );

void    DCPU_DumpCmdFifo(       HWND                    hWnd,
                                FILE                    *pFile,
                                BOOL                    bError );

void    DCPU_DumpCmdFifoToFile( FILE                    *pFile );

void    DCPU_ClearCmdTag(       pDCC_CMD_TAG            pCmdTag );

void    DCPU_CopySndCommand(    SND_CMD_MNEMONIC        nCommand,
                                int                     nNbArgs, 
                                int                     *pDataArray,
                                DCC_COMMAND             *pCommand );

BOOL    DCPU_IsExpectedResponse( RCV_CMD_MNEMONIC       nRdCommand,
                                DCC_COMMAND             *pRcvCommand, 
                                DCC_COMMAND             *pSndCommand );

void    DCPU_ParseCommandCtxt( 
                pDCC_COMMAND_CTXT       pCmdCtxt,
                RCV_CMD_MNEMONIC        *pCommand,
                int                     *pAddress,
                int                     xData[MAX_NB_DCC_DATA_WORDS],
                int                     *pNbBytes,
                unsigned char           xByte[MAX_NB_CMD_DATA] );

void    DCPU_MapResponseToGeneric( 
                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                pDCC_COMMAND_CTXT       pCmdCtxt,
                RCV_CMD_MNEMONIC        *pCommand );

void    DCPU_MapMessageToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData,
                                pDCC_COMMAND_CTXT       pCmdCtxt,
                                RCV_CMD_MNEMONIC        *pCommand );

void    DCPU_MapFeedbackBroadcastToGeneric( 
                                pDGI_GENERIC_RCV_DATA   pGenericRcvData );

int     DCPU_F0F12Byte2Int(     unsigned char           cFuncA, 
                                unsigned char           cFuncB );


/* DCC_COMMAND_CTXT HANDLING */

pDCC_COMMAND_CTXT 
        DCPU_CreateCommandCtxt( void );

void    DCPU_ClearCommandCtxt(  pDCC_COMMAND_CTXT       pCmdCtxt );

void    DCPU_CopyCommanCtxt(    pDCC_COMMAND_CTXT       pDstCmdCtxt, 
                                pDCC_COMMAND_CTXT       pSrcCmdCtxt );

void    DCPU_ClearCommand(      pDCC_COMMAND            pCommand );

pDCC_COMMAND_CTXT
        DCPU_NewCommandCtxt(    pDCC_CMD_CTXT_ENV       pCmdEnv,
                                CMD_PRIORITY            nPriority );

void    DCPU_DetachCommandCtxt( pDCC_CMD_CTXT_ENV       pCmdEnv );

void    DCPU_FreeCommandCtxtList( pDCC_CMD_CTXT_ENV     pCmdEnv );

/* DETECTOR ARRAY HANDLING FUNCTIONS */

BOOL    DCPU_UpdateDetectorArray( DCC_COMMAND_CTXT      *pCmdCtxt );

/* DCC FORMAT CONVERSION UTILITIES */

int     DCPU_F0F12Byte2Int(     unsigned char           cFuncA, 
                                unsigned char           cFuncB );

unsigned char   
        DCPU_SpeedInt2Byte(     int                     nSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode );

int     DCPU_SpeedByte2Int(     unsigned char           cSpeed, 
                                SPEED_STEP_MODE         nSpeedStepMode,
                                BOOL                    bReverse );

int     DCPU_AddressByte2Int(   unsigned char           cADH, 
                                unsigned char           cADL );

/* DATA COM UTILITIES */

void    DCPU_SetTimeOut(        HANDLE                  hCom, 
                                DWORD                   nTimeOut );

/* STRING UTILITIES */

BOOL    DCPU_SkipBlanks(        unsigned char           xBuffer[], 
                                int                     *pIndex );

BOOL    DCPU_GetNextCharIndex(  unsigned char           xBuffer[], 
                                unsigned char           xExpectedChar[],
                                int                     *pIndex );

BOOL    DCPU_IsDecimal(         unsigned char           cChar );

BOOL    DCPU_IsHexaDecimal(     unsigned char           cChar );

BOOL    DCPU_ExtractDigits(     unsigned char           xRcvBuffer[], 
                                int                     *pIndex, 
                                int                     nNbDigits, 
                                unsigned char           xDigitArray[] );

unsigned char
        DCPU_SetByte(           unsigned char           cHigh, 
                                unsigned char           cLow );

BOOL    DCPU_Ascii2Hexa(        unsigned char           *pBuffer, 
                                int                     nNbChars, 
                                unsigned int            *pValue );


unsigned char *
        DCPU_DisplayVersion(    unsigned char           xByte[], 
                                int                     nNbDigits, 
                                int                     nAfterDot,
                                unsigned char           xBuffer[] );

BOOL    DCPU_IsKnownErrorCode(  unsigned char           cByte, 
                                int                     *pErrorIdx );

BOOL    DCPU_CommandStationBusy( RD_ERROR               nRdError );

void    DCPU_PrintDebugMessage( char                    *sDbgMsg );
